#  2023-01-27 Cornelius Kölbel <cornelius@privacyidea.org>
#             Create this module for enabling decorators for API calls
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
These are logical policy functions that are usually used in policy API decorators, but
in some cases also used beside the API.
Like policies, that are supposed to read and pass parameters during enrollment of a token.
"""
from dataclasses import dataclass
import logging

from privacyidea.lib.container import get_container_realms, find_container_for_token, find_container_by_serial
from privacyidea.lib.log import log_with
from privacyidea.lib.policy import Match, SCOPE, ACTION
from privacyidea.lib.error import PolicyError, ResourceNotFoundError
from privacyidea.lib.token import get_tokens_from_serial_or_user, get_token_owner, get_realms_of_token
from privacyidea.lib.user import User

log = logging.getLogger(__name__)


@dataclass
class UserAttributes:
    role: str = None
    username: str = None
    realm: str = None
    resolver: str = None
    adminuser: str = None
    adminrealm: str = None
    additional_realms: list = None


@log_with(log)
def get_init_tokenlabel_parameters(g, params=None, token_type="hotp", user_object=None):
    """
    This helper function modifies the request parameters in regards
    to enrollment policies tokenlabel, tokenissuer, appimage, force_app_pin

    :param params: The request parameter
    :param user_object: User object in the request
    :return: modified request parameters
    """
    params = params or {}
    label_pols = Match.user(g, scope=SCOPE.ENROLL, action=ACTION.TOKENLABEL,
                            user_object=user_object).action_values(unique=True, allow_white_space_in_action=True)
    if len(label_pols) == 1:
        # The policy was set, so we need to set the tokenlabel in the request.
        params[ACTION.TOKENLABEL] = list(label_pols)[0]

    issuer_pols = Match.user(g, scope=SCOPE.ENROLL, action=ACTION.TOKENISSUER,
                             user_object=user_object).action_values(unique=True, allow_white_space_in_action=True)
    if len(issuer_pols) == 1:
        params[ACTION.TOKENISSUER] = list(issuer_pols)[0]

    imageurl_pols = Match.user(g, scope=SCOPE.ENROLL, action=ACTION.APPIMAGEURL,
                               user_object=user_object).action_values(unique=True, allow_white_space_in_action=True)
    if len(imageurl_pols) == 1:
        params[ACTION.APPIMAGEURL] = list(imageurl_pols)[0]

    # check the force_app_pin policy
    app_pin_pols = Match.user(g, scope=SCOPE.ENROLL,
                              action='{0!s}_{1!s}'.format(token_type, ACTION.FORCE_APP_PIN),
                              user_object=user_object).any()
    if app_pin_pols:
        params[ACTION.FORCE_APP_PIN] = True

    return params


def get_pushtoken_add_config(g, params=None, user_obj=None):
    """
    This helper function modifies the request parameters in regards
    to enrollment policies for push tokens.

    :param params: The request parameter
    :param user_object: User object in the request
    :return: modified request parameters
    """
    params = params or {}
    from privacyidea.lib.tokens.pushtoken import PUSH_ACTION

    # Get the firebase configuration from the policies
    firebase_config = Match.user(g, scope=SCOPE.ENROLL,
                                 action=PUSH_ACTION.FIREBASE_CONFIG,
                                 user_object=user_obj if user_obj else None
                                 ).action_values(unique=True,
                                                 allow_white_space_in_action=True)
    if len(firebase_config) == 1:
        params[PUSH_ACTION.FIREBASE_CONFIG] = list(firebase_config)[0]
    else:
        raise PolicyError("Missing enrollment policy for push token: {0!s}".format(PUSH_ACTION.FIREBASE_CONFIG))

    # Get the sslverify definition from the policies
    ssl_verify = Match.user(g, scope=SCOPE.ENROLL, action=PUSH_ACTION.SSL_VERIFY,
                            user_object=user_obj if user_obj else None).action_values(unique=True)
    if len(ssl_verify) == 1:
        params[PUSH_ACTION.SSL_VERIFY] = list(ssl_verify)[0]
    else:
        params[PUSH_ACTION.SSL_VERIFY] = "1"

    # Get the TTL and the registration URL from the policies
    registration_url = Match.user(g, scope=SCOPE.ENROLL, action=PUSH_ACTION.REGISTRATION_URL,
                                  user_object=user_obj if user_obj else None) \
        .action_values(unique=True, allow_white_space_in_action=True)
    if len(registration_url) == 1:
        params[PUSH_ACTION.REGISTRATION_URL] = list(registration_url)[0]
    else:
        raise PolicyError("Missing enrollment policy for push token: {0!s}".format(PUSH_ACTION.REGISTRATION_URL))
    ttl = Match.user(g, scope=SCOPE.ENROLL, action=PUSH_ACTION.TTL,
                     user_object=user_obj if user_obj else None) \
        .action_values(unique=True, allow_white_space_in_action=True)
    if len(ttl) == 1:
        params[PUSH_ACTION.TTL] = list(ttl)[0]
    else:
        params[PUSH_ACTION.TTL] = "10"
    return params


def get_token_user_attributes(serial: str):
    """
    Get the user attributes from the token owner and the token realms.

    :param serial: The serial of the token
    :return: UserAttributes dataclass
    """
    user_attributes = UserAttributes()
    # get user attributes from the token
    try:
        token = get_tokens_from_serial_or_user(serial, user=None)[0]
        token_owner = get_token_owner(serial)
    except ResourceNotFoundError:
        token = None
        token_owner = None
        log.error(f"Could not find token with serial {serial}.")
    if token_owner:
        user_attributes.username = token_owner.login
        user_attributes.realm = token_owner.realm
        user_attributes.resolver = token_owner.resolver
    if token:
        user_attributes.additional_realms = token.get_realms()
    return user_attributes


def get_container_user_attributes(container_serial: str) -> UserAttributes:
    """
    Get the user and container realms from the request.
    If a user attribute (username, realm, resolver) is not available, it is set to None.
    If no container realms are available an empty list is set.

    :param container_serial: Serial of the container
    :return: container_owner_attributes
    """
    container_owner_attributes = UserAttributes()

    try:
        container = find_container_by_serial(container_serial)
    except ResourceNotFoundError:
        container = None
        log.error(f"Could not find container with serial {container_serial}.")
    if container:
        container_owners = container.get_users()
        container_owner = container_owners[0] if container_owners else None
        if container_owner:
            container_owner_attributes.username = container_owner.login
            container_owner_attributes.realm = container_owner.realm
            container_owner_attributes.resolver = container_owner.resolver
        container_owner_attributes.additional_realms = [realm.name for realm in container.realms]

    return container_owner_attributes


def check_token_action_allowed(g, action: str, serial: str, user_attributes: UserAttributes) -> bool:
    """
    Retrieves user attributes from the token and checks if the logged-in user is allowed to perform the action on the
    token.

    For admins, the policies either need to match the token owner or at least one of the token realms.
    If no user attributes (username, realm, resolver) are available, the policies are filtered for generic policies
    without conditions on the user. Only for the action ASSIGN, all policies are considered, ignoring the username,
    realm, and resolver conditions. The token realms are still taken into account. This shall allow helpdesk admins
    to assign their users to tokens without owner.

    :param g: The global flask object g
    :param action: The action to be performed on the token
    :param serial: The serial of the token
    :param user_attributes: User attributes of the logged-in user
    :return: True if the action is allowed, False otherwise
    """
    if serial:
        token_owner_attributes = get_token_user_attributes(serial)
    else:
        token_owner_attributes = UserAttributes()

    if user_attributes.role == "admin":
        if action == ACTION.ASSIGN:
            # Assigning a user to a token is only possible if the token has no owner yet.
            # To avoid helpdesk admins (for a specific resolver) lose access on their tokens while changing the owner
            # of a token, they are allowed to assign their users to tokens without owner.
            # Note: the policies are still filtered by the token realms.
            user_attributes.username = token_owner_attributes.username or None
            user_attributes.realm = token_owner_attributes.realm or None
            user_attributes.resolver = token_owner_attributes.resolver or None
        else:
            # If no user is available, explicitly filter for generic policies without conditions on the user
            user_attributes.username = token_owner_attributes.username or ""
            user_attributes.realm = token_owner_attributes.realm or ""
            user_attributes.resolver = token_owner_attributes.resolver or ""
        user_attributes.additional_realms = token_owner_attributes.additional_realms or None
    elif user_attributes.role == "user" and serial:
        # for adding / removing tokens from a container, the user has to be the owner of the token
        if action in [ACTION.CONTAINER_ADD_TOKEN, ACTION.CONTAINER_REMOVE_TOKEN]:
            if not user_is_owner(user_attributes, token_owner_attributes, allow_no_owner=False):
                return False

    # Check action for the token
    action_allowed = Match.generic(g,
                                   scope=user_attributes.role,
                                   action=action,
                                   user=user_attributes.username,
                                   resolver=user_attributes.resolver,
                                   realm=user_attributes.realm,
                                   adminrealm=user_attributes.adminrealm,
                                   adminuser=user_attributes.adminuser,
                                   additional_realms=user_attributes.additional_realms).allowed()

    if action_allowed and action == ACTION.CONTAINER_ADD_TOKEN:
        # Adding a token to a container will remove it from the old container: Check if the remove action is allowed
        try:
            old_container = find_container_for_token(serial)
        except ResourceNotFoundError:
            old_container = None

        if old_container:
            action_allowed = check_container_action_allowed(g, ACTION.CONTAINER_REMOVE_TOKEN, old_container.serial,
                                                            user_attributes)
            if not action_allowed:
                log.info(f"Token {serial} is in container {old_container.serial}. The user is not allowed to remove the"
                         " token from this container.")

    return action_allowed


def check_container_action_allowed(g, action: str, container_serial: str, user_attributes: UserAttributes) -> bool:
    """
    Retrieves user attributes from the container and checks if the logged-in user is allowed to perform the action
    on the container.

    For admins, the policies either need to match the container owner or at least one of the container realms.
    If no user attributes (username, realm, resolver) are available, the policies are filtered for generic policies
    without conditions on the user. Only for the action CONTAINER_ASSIGN_USER, all policies are considered, ignoring
    the username, realm, and resolver conditions. The container realms are still taken into account. This shall allow
    helpdesk admins to assign their users to containers without owner.

    For the action CONTAINER_CREATE, the user attributes from the parameters are considered, as the container has no
    owner yet.

    :param g: The global flask object g
    :param action: The action to be performed on the container
    :param container_serial: The serial of the container
    :param user_attributes: User attributes of the logged-in user
    :return: True if the action is allowed, False otherwise
    """
    user_attributes.additional_realms = None
    container_owner_attributes = UserAttributes()
    if container_serial:
        # get user attributes from the container
        container_owner_attributes = get_container_user_attributes(container_serial)

    if user_attributes.role == "admin":
        if action == ACTION.CONTAINER_ASSIGN_USER:
            # Assigning a user to a container is only possible if the container has no owner yet.
            # To avoid helpdesk admins (for a specific resolver) lose access on their containers while changing the
            # owner of a container, they are allowed to assign their users to containers without user.
            # Note: the policies are still filtered by the container realms.
            user_attributes.username = container_owner_attributes.username or None
            user_attributes.realm = container_owner_attributes.realm or None
            user_attributes.resolver = container_owner_attributes.resolver or None
        elif action == ACTION.CONTAINER_CREATE:
            # If the container is created, it has no owner yet, instead check the user attributes from the parameters
            user_attributes.username = user_attributes.username or ""
            user_attributes.realm = user_attributes.realm or ""
            user_attributes.resolver = user_attributes.resolver or ""
        else:
            # If no user is available, explicitly filter for generic policies without conditions on the user
            user_attributes.username = container_owner_attributes.username or ""
            user_attributes.realm = container_owner_attributes.realm or ""
            user_attributes.resolver = container_owner_attributes.resolver or ""
        user_attributes.additional_realms = container_owner_attributes.additional_realms or None
    elif user_attributes.role == "user" and container_serial:
        # check if the user is the owner of the container
        if action == ACTION.CONTAINER_CREATE:
            # container has no owner yet, skip this check
            is_owner = True
        elif action == ACTION.CONTAINER_ASSIGN_USER:
            # users are allowed to assign containers without owner
            is_owner = user_is_owner(user_attributes, container_owner_attributes, allow_no_owner=True)
        else:
            is_owner = user_is_owner(user_attributes, container_owner_attributes, allow_no_owner=False)
        if not is_owner:
            return False

    # Check action for container
    action_allowed = Match.generic(g,
                                   scope=user_attributes.role,
                                   action=action,
                                   user=user_attributes.username,
                                   resolver=user_attributes.resolver,
                                   realm=user_attributes.realm,
                                   adminrealm=user_attributes.adminrealm,
                                   adminuser=user_attributes.adminuser,
                                   additional_realms=user_attributes.additional_realms).allowed()
    return action_allowed


def user_is_owner(logged_in_user_attr: UserAttributes, owner_attr: UserAttributes,
                  allow_no_owner: bool = False) -> bool:
    """
    Checks if the logged-in user is the owner of the token or container.

    :param logged_in_user_attr: User attributes of the logged-in user
    :param owner_attr: User attributes of the token or container owner
    :param allow_no_owner: If True, the user is considered the owner if the owner is None
    :return: True if the user is the owner, False otherwise
    """
    user = User(logged_in_user_attr.username, logged_in_user_attr.realm, logged_in_user_attr.resolver)
    owner = User(owner_attr.username, owner_attr.realm, owner_attr.resolver)
    is_owner = user == owner
    if not is_owner and allow_no_owner and not owner:
        is_owner = True
    return is_owner



from privacyidea.models import MonitoringStats, db
from privacyidea.lib.monitoringstats import (write_stats, delete_stats,
                                             get_stats_keys, get_values,
                                             get_last_value)

from .base import MyTestCase
import datetime
from dateutil.tz import tzlocal, tzutc
from datetime import timedelta


class TokenModelTestCase(MyTestCase):

    def test_01_write_stats(self):
        key1 = "some_key"
        write_stats(key1, 12)
        write_stats(key1, 13)

        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 2)
        # Assert naive datetime
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).first().timestamp.tzinfo, None)

        # Now we write a new value, but with the parameter to delete old values
        write_stats(key1, 14, reset_values=True)
        # Since the sqlstats monitoring module handles the db in its own
        # session, we need to refresh the current session in order to retrieve
        # the updated (removed) data
        db.session.commit()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 1)

    def test_02_delete_stats(self):
        key1 = "otherkey"
        now = datetime.datetime.now(tzlocal())
        write_stats(key1, 12, timestamp=now- timedelta(days=1))
        write_stats(key1, 13, timestamp=now)
        write_stats(key1, 14, timestamp=now + timedelta(days=1))
        db.session.commit()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 3)

        # delete the last two entries
        r = delete_stats(key1, start_timestamp=now - timedelta(minutes=60))

        # check there is only one entry
        db.session.commit()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 1)
        self.assertEqual(r, 2)

        # Again write three entries
        write_stats(key1, 13, timestamp=now)
        write_stats(key1, 14, timestamp=now + timedelta(days=1))
        db.session.commit()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 3)

        # Delete the first two entries
        r = delete_stats(key1, end_timestamp=now + timedelta(minutes=60))
        db.session.commit()
        self.assertEqual(MonitoringStats.query.filter_by(stats_key=key1).count(), 1)
        self.assertEqual(r, 2)

    def test_03_get_keys(self):
        # delete old entries
        keys = get_stats_keys()
        for k in keys:
            delete_stats(k)

        # write new stats entries
        write_stats("key1", 13)
        write_stats("key1", 13)
        write_stats("key1", 13)
        write_stats("key2", 12)
        write_stats("key3", 12)

        db.session.commit()
        keys = get_stats_keys()
        self.assertEqual(len(keys), 3)
        self.assertTrue("key1" in keys)
        self.assertTrue("key2" in keys)
        self.assertTrue("key3" in keys)

    def test_04_get_values(self):
        # delete old entries
        keys = get_stats_keys()
        for k in keys:
            delete_stats(k)

        ts = datetime.datetime.now(tzlocal())
        write_stats("key1", 1, timestamp=ts - timedelta(minutes=10))
        write_stats("key1", 2, timestamp=ts - timedelta(minutes=9))
        write_stats("key1", 3, timestamp=ts - timedelta(minutes=8))
        write_stats("key1", 4, timestamp=ts - timedelta(minutes=7))
        write_stats("key1", 5, timestamp=ts - timedelta(minutes=6))
        write_stats("key1", 6, timestamp=ts - timedelta(minutes=5))
        write_stats("key1", 7, timestamp=ts - timedelta(minutes=4))
        write_stats("key1", 8, timestamp=ts - timedelta(minutes=3))
        write_stats("key1", 9, timestamp=ts - timedelta(minutes=2))
        write_stats("key1", 10, timestamp=ts - timedelta(minutes=1))

        db.session.commit()
        r = get_values("key1")
        self.assertEqual(len(r), 10)
        # The third entry is a 3
        self.assertEqual(r[2][1], 3)
        # The last value is a 10
        self.assertEqual(r[9][1], 10)

        r = get_values("key1",
                       start_timestamp=ts - timedelta(minutes=8),
                       end_timestamp=ts - timedelta(minutes=4))
        # We get 3,4,5,6,7
        self.assertEqual([entry[1] for entry in r], [3, 4, 5, 6, 7])
        # Assert it is the correct time, and timezone-aware UTC
        self.assertEqual(r[0][0], ts - timedelta(minutes=8))
        self.assertEqual(r[0][0].tzinfo, tzutc())
        self.assertEqual(r[-1][0], ts - timedelta(minutes=4))

        r = get_values("key1",
                       start_timestamp=ts - timedelta(minutes=8))
        self.assertEqual([entry[1] for entry in r], [3, 4, 5, 6, 7, 8, 9, 10])

        r = get_values("key1",
                       end_timestamp=ts - timedelta(minutes=8))
        self.assertEqual([entry[1] for entry in r], [1, 2, 3])

        # Get the last value of key1
        r = get_last_value("key1")
        self.assertEqual(r, 10)


#  2020-02-27 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial write. Allow tokens to be initialized with an otpkey
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This is the implementation of an Indexed Secret-Token.
It is a challenge response token, that asks the user for certain positions
of the secret string.
The user must know the secret and return the characters from the requested positions.

The secret is stored in the encrypted OTP KEY.

The code is tested in tests/test_lib_tokens_indexedsecret

Note that this token has no "check_otp" function. So it is not even
possible to do a "single shot" authentication, since the base class check_otp always
returns -1.
"""

import logging
import datetime
from privacyidea.lib.tokenclass import TokenClass, AUTHENTICATIONMODE
from privacyidea.lib.policy import (SCOPE, ACTION, GROUP, comma_escape_text,
                                    get_action_values_from_options)
from privacyidea.lib.crypto import urandom, safe_compare
from privacyidea.lib.log import log_with
from privacyidea.lib import _, lazy_gettext
from privacyidea.lib.utils import to_unicode
from privacyidea.lib.challenge import get_challenges
from privacyidea.models import Challenge
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.error import ValidateError

log = logging.getLogger(__name__)

DEFAULT_CHALLENGE_TEXT = lazy_gettext("Please enter the positions {0!s} from your secret.")
DEFAULT_POSITION_COUNT = 2


class PIIXACTION(object):
    COUNT = "count"
    PRESET_ATTRIBUTE = "preset_attribute"
    FORCE_ATTRIBUTE = "force_attribute"


class IndexedSecretTokenClass(TokenClass):
    """
    Implementation of the Indexed Secret Token Class, that asks the user for certain
    positions in a shared secret.
    """
    mode = [AUTHENTICATIONMODE.CHALLENGE]

    # The token type provides means to verify the enrollment
    can_verify_enrollment = True

    def __init__(self, aToken):
        TokenClass.__init__(self, aToken)
        self.set_type("indexedsecret")

    @staticmethod
    def get_class_type():
        """
        return the generic token class identifier
        """
        return "indexedsecret"

    @staticmethod
    def get_class_prefix():
        return "PIIX"

    @staticmethod
    def get_class_info(key=None, ret='all'):
        """
        returns all or a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined

        :return: subsection if key exists or user defined
        :rtype : s.o.
        """
        res = {'type': 'indexedsecret',
               'title': _('Indexed Secret Token'),
               'description':
                   _('IndexedSecret: Request certain positions of a shared secret from the user.'),
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {SCOPE.AUTH: {
                   ACTION.CHALLENGETEXT: {
                       'type': 'str',
                       'group': "Indexed Secret Token",
                       'desc': _('Use an alternative challenge text for telling the '
                                 'user which positions of the secret he should enter. You can also '
                                 'use tags for automated replacement. Check out the documentation '
                                 'for more details.')
                               + " " + comma_escape_text
                   },
                   PIIXACTION.COUNT: {
                       'type': 'int',
                       'desc': _('Number of necessary positions to be answered by the user.'),
                       'group': "Indexed Secret Token"
                   }
               },
                   SCOPE.WEBUI: {
                       PIIXACTION.PRESET_ATTRIBUTE: {
                           'type': 'str',
                           'desc': _("Preset the enrollment with the value of the given attribute."),
                           'group': "token"
                       }
                   },
                   SCOPE.USER: {
                       PIIXACTION.FORCE_ATTRIBUTE: {
                           'type': 'str',
                           'desc': _("The attribute whose value should be force set during enrollment."),
                           'group': "enrollment"
                       }
                   },
                   SCOPE.ADMIN: {
                       PIIXACTION.FORCE_ATTRIBUTE: {
                           'type': 'str',
                           'desc': _("The attribute whose value should be force set during enrollment."),
                           'group': "enrollment"
                       }
                   },
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of indexed secret tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active indexed secret"
                                     " tokens assigned."),
                           'group': GROUP.TOKEN
                   }
               }
           }
        }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res

        return ret

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        update - process initialization parameters

        :param param: dict of initialization parameters
        :type param: dict

        :return: nothing

        """
        if 'verify' not in param and 'genkey' not in param and 'otpkey' not in param:
            param['genkey'] = 1

        TokenClass.update(self, param, reset_failcount)

        return

    @log_with(log)
    def create_challenge(self, transactionid=None, options=None):
        """
        create a challenge, which is submitted to the user

        :param transactionid: the id of this challenge
        :param options: the request context parameters / data
        :return: tuple of (bool, message, transactionid, attributes)
        :rtype: tuple

        The return tuple builds up like this:
        ``bool`` if submit was successful;
        ``message`` which is displayed in the JSON response;
        additional challenge ``reply_dict``, which are displayed in the JSON challenges response.
        """
        options = options or {}
        return_message = get_action_values_from_options(SCOPE.AUTH,
                                                        "{0!s}_{1!s}".format(self.get_class_type(),
                                                                             ACTION.CHALLENGETEXT),
                                                        options) or DEFAULT_CHALLENGE_TEXT

        return_message = return_message.replace(r'\,', ',')

        if self.get_tokeninfo("multichallenge"):
            # In case of multichallenge we ask only once.
            position_count = 1
        else:
            position_count = int(get_action_values_from_options(SCOPE.AUTH,
                                                                "{0!s}_{1!s}".format(self.get_class_type(),
                                                                                     PIIXACTION.COUNT),
                                                                options) or DEFAULT_POSITION_COUNT)

        attributes = {'state': transactionid}
        validity = 120

        if self.is_active() is True:
            # We need to get a number of random positions from the secret string
            secret_length = len(self.token.get_otpkey().getKey())
            if not secret_length:
                raise ValidateError("The indexedsecret token has an empty secret and "
                                    "can not be used for authentication.")
            random_positions = [urandom.randint(1, secret_length) for _x in range(0, position_count)]
            position_str = ",".join(["{0!s}".format(x) for x in random_positions])
            attributes["random_positions"] = random_positions

            db_challenge = Challenge(self.token.serial,
                                     transaction_id=transactionid,
                                     challenge=options.get("challenge"),
                                     data=position_str,
                                     session=options.get("session"),
                                     validitytime=validity)
            db_challenge.save()
            transactionid = transactionid or db_challenge.transaction_id
            return_message = return_message.format(position_str)

        expiry_date = datetime.datetime.now() + datetime.timedelta(seconds=validity)
        attributes['valid_until'] = "{0!s}".format(expiry_date)
        reply_dict = {"attributes": attributes}

        return True, return_message, transactionid, reply_dict

    @check_token_locked
    def check_challenge_response(self, user=None, passw=None, options=None):
        """
        This method verifies if there is a matching challenge for the given
        passw and also verifies if the response is correct.

        It then returns 1 in case of success
        In case of failure it returns -1

        :param user: the requesting user
        :type user: User object
        :param passw: the password (pin+otp)
        :type passw: string
        :param options: additional arguments from the request, which could
                        be token specific. Usually "transactionid"
        :type options: dict
        :return: return success
        :rtype: int
        """
        options = options or {}
        r_success = -1

        # fetch the transaction_id
        transaction_id = options.get('transaction_id')
        if transaction_id is None:
            transaction_id = options.get('state')

        # get the challenges for this transaction ID
        if transaction_id is not None:
            challengeobject_list = get_challenges(serial=self.token.serial,
                                                  transaction_id=transaction_id)

            for challengeobject in challengeobject_list:

                if challengeobject.is_valid():
                    # challenge is still valid
                    # Add the challenge to the options for check_otp
                    options["challenge"] = challengeobject.challenge
                    options["data"] = [int(c) for c in challengeobject.data.split(",")]
                    # Now see if the answer is the right indexes
                    secret_string = to_unicode(self.token.get_otpkey().getKey())
                    if len(options["data"]) == len(passw):
                        expected_answer = "".join([secret_string[x - 1] for x in options["data"]])
                        if safe_compare(passw, expected_answer):
                            r_success = 1
                            # Set valid OTP to true. We must not delete the challenge now,
                            # Since we need it for further mutlichallenges
                            challengeobject.set_otp_status(True)
                            log.debug("The presented answer was correct.")
                            break
                        else:
                            log.debug("The presented answer was wrong.")
                            # increase the received_count
                            challengeobject.set_otp_status()
                    else:
                        log.info("Length of password does not match the requested number of positions.")
                        # increase the received_count
                        challengeobject.set_otp_status()

        self.challenge_janitor()
        return r_success

    @log_with(log)
    def is_challenge_request(self, passw, user=None, options=None):
        """
        check, if the request would start a challenge

        We need to define the function again, to get rid of the
        is_challenge_request-decorator of the HOTP-Token

        :param passw: password, which might be pin or pin+otp
        :param options: dictionary of additional request parameters

        :return: returns true or false
        """
        return self.check_pin(passw, user=user, options=options)

    @log_with(log)
    def has_further_challenge(self, options=None):
        """
        Check if we should do multi challenge at all and
        then if there are further positions to query.

        :param options: Options dict
        :return: True, if further challenge is required.
        """
        if self.get_tokeninfo("multichallenge"):
            transaction_id = options.get('transaction_id')
            challengeobject_list = get_challenges(serial=self.token.serial,
                                                  transaction_id=transaction_id)

            position_count = int(get_action_values_from_options(SCOPE.AUTH,
                                                                "{0!s}_{1!s}".format(self.get_class_type(),
                                                                                     PIIXACTION.COUNT),
                                                                options) or DEFAULT_POSITION_COUNT)
            if len(challengeobject_list) == 1:
                session = int(challengeobject_list[0].session or "0") + 1
                options["session"] = "{0!s}".format(session)
                if session < position_count:
                    return True

        return False

    def prepare_verify_enrollment(self, options=None):
        """
        This is called, if the token should be enrolled in a way, that the user
        needs to provide a proof, that the server can verify, that the token
        was successfully enrolled. A challenge for the indexed secret token is created
        and the user will later have to answer this.

        The returned dictionary is added to the response in "detail" -> "verify".

        :return: A dictionary with information that is needed to trigger the verification.
        """
        _, return_message, transaction_id, reply_dict = self.create_challenge(options=options)
        return {"message": return_message}

    def verify_enrollment(self, verify):
        """
        This is called during the 2nd step of the verified enrollment.
        This method verifies the actual response from the user.
        Returns true, if the verification was successful.

        :param verify: The response given by the user
        :return: True
        """
        # During the enrollment one challenge has been created. The token is not fit
        # for authentication, yet. So there should only be one challenge for this token
        # in the challenge table. Find it!
        chals = get_challenges(serial=self.token.serial)
        if len(chals) != 1:  # pragma: no cover
            log.error("Something is wrong. There is more than one challenge!")
        transaction_id = chals[0].transaction_id
        r = self.check_challenge_response(passw=verify,
                                          options={"transaction_id": transaction_id})
        log.debug("Enrollment verified: {0!s}".format(r))
        return r >= 0



# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""CLI commands for configuring the privacyIDEA server"""
import copy
import sys
import ast
from functools import partial
import click
from flask import current_app
from flask.cli import AppGroup
import json
import traceback
import yaml

from privacyidea.lib.authcache import cleanup
from privacyidea.lib.caconnector import (get_caconnector_list,
                                         get_caconnector_class,
                                         get_caconnector_object,
                                         save_caconnector)
from privacyidea.lib.caconnectors.localca import ATTR
from privacyidea.lib.crypto import create_hsm_object
from privacyidea.lib.error import ResourceNotFoundError
from privacyidea.lib.event import EventConfiguration, enable_event, delete_event
from privacyidea.lib.policy import (PolicyClass, enable_policy, delete_policy,
                                    set_policy)
from privacyidea.lib.realm import (get_realms, delete_realm, set_default_realm,
                                   get_default_realm)
from privacyidea.lib.resolver import (save_resolver, get_resolver_list)
from privacyidea.lib.utils import get_version_number
from privacyidea.lib.utils.export import EXPORT_FUNCTIONS, IMPORT_FUNCTIONS
from privacyidea.cli.pimanage.challenge import challenge_cli

config_cli = AppGroup("config", help="Manage the privacyIDEA server configuration")

config_cli.add_command(challenge_cli)

ca_cli = AppGroup("ca", help="Manage Certificate Authorities")


@ca_cli.command("list")
@click.option("-v", "--verbose", is_flag=True,
              help="Additionally output the configuration of the CA connector")
def ca_list(verbose):
    """
    List the Certificate Authorities.
    """
    lst = get_caconnector_list()
    for ca in lst:
        click.echo(f"{ca.get('connectorname')} (type {ca.get('type')})")
        if verbose:
            for (k, v) in ca.get("data").items():
                click.echo(f"\t{k:20}: {v}")


@ca_cli.command("create", short_help="Create a new CA connector.")
@click.argument('name', type=str)
@click.option('-t', '--type', "catype", default="local",
              help='The type of the new CA', show_default=True)
@click.pass_context
def ca_create(ctx, name, catype):
    """
    Create a new CA connector identified by NAME.

    If the type "local" is given, also the directory
    structure, the openssl.cnf and the CA key pair will be created.
    """
    ca = get_caconnector_object(name)
    if ca:
        click.secho(f"A CA connector with the name '{name}' already exists.",
                    fg="red")
        ctx.exit(1)
    click.echo(f"Creating CA connector of type {catype}.")
    ca_class = get_caconnector_class(catype)
    if not ca_class:
        click.secho(f"Unknown CA type {catype}.", fg="red")
        ctx.exit(1)
    ca_params = ca_class.create_ca(name)
    r = save_caconnector(ca_params)
    if r:
        click.secho(f"Saved CA Connector with ID {r}.", fg="green")
        if ca_params["type"] == "local":
            click.secho(f"Warning: Be sure to set the correct access rights "
                        f"to the directory {ca_params[ATTR.WORKING_DIR]}.", fg="yellow")
    else:
        click.secho(f"Error saving CA connector {name}.", fg="red")


@ca_cli.command("create_crl", short_help="Create and publish CRL")
@click.argument("name", type=str)
@click.option("-f", "--force", is_flag=True,
              help="Enforce creation of a new CRL")
def ca_create_crl(name, force):
    """
    Create and publish the CRL for the CA NAME.
    """
    ca_obj = get_caconnector_object(name)
    r = ca_obj.create_crl(check_validity=not force)
    if not r:
        click.secho("The CRL was not created.", fg="yellow")
    else:
        click.echo(f"The CRL {r} was created.")


config_cli.add_command(ca_cli)
hsm_cli = AppGroup("hsm", help="Manage hardware security modules")


@hsm_cli.command("create_keys")
def hsm_create_keys():
    """
    Create new encryption keys on the HSM.
    Be sure to first set up the HSM module, the PKCS11
    module and the slot/password for the given HSM in your pi.cfg.
    Set the variables PI_HSM_MODULE, PI_HSM_MODULE_MODULE, PI_HSM_MODULE_SLOT
    and PI_HSM_MODULE_PASSWORD.
    """
    hsm_object = create_hsm_object(current_app.config)
    r = hsm_object.create_keys()
    click.echo("Please add the following to your pi.cfg:")
    click.echo(f"PI_HSM_MODULE_KEY_LABEL_TOKEN = {r.get('token')!r}")
    click.echo(f"PI_HSM_MODULE_KEY_LABEL_CONFIG = {r.get('config')!r}")
    click.echo(f"PI_HSM_MODULE_KEY_LABEL_VALUE = {r.get('value')!r}")


config_cli.add_command(hsm_cli)
realm_cli = AppGroup("realm", help="Manage realms")


@realm_cli.command("list")
def realm_list():
    """
    list the available realms.
    The '*' denotes the default realm.
    """
    realm_lst = get_realms()
    def_realm = get_default_realm()
    for (name, realm_data) in realm_lst.items():
        if name == def_realm:
            name = f"* {name}"
        resolver_names = [x.get("name") for x in realm_data.get("resolver")]
        click.echo(f"{name:16}: {resolver_names}")


@realm_cli.command("create")
@click.argument("name")
@click.argument("resolvers", nargs=-1)
def realm_create(name, resolvers):
    """
    Create a new realm.
    This command will create a new realm with the given resolvers.

    An existing realm with the same name will be replaced.
    """
    from privacyidea.lib.realm import set_realm
    (added, failed) = set_realm(name, [{'name': res} for res in resolvers])
    if failed:
        click.secho(f"Realm '{name}' created. Following resolvers could not be "
                    f"assigned: {failed}", fg="yellow")
    else:
        click.secho(f"Successfully created realm '{name}' with resolver: {added}.",
                    fg="green")


@realm_cli.command("delete")
@click.argument("realm", type=str)
def realm_delete(realm):
    """
    Delete the given REALM
    """
    try:
        delete_realm(realm)
    except ResourceNotFoundError as e:
        click.secho(f"Could not delete realm '{realm}': {e!r}", fg="red")
    else:
        click.secho(f"Realm '{realm}' successfully deleted.", fg="green")


@realm_cli.command("set_default")
@click.argument("realm", type=str)
def realm_set_default(realm):
    """
    Set the given REALM as the default realm
    """
    try:
        set_default_realm(realm)
    except ResourceNotFoundError as e:
        click.secho(f"Could not set realm {realm} as default: {e!r}", fg="red")
    else:
        click.secho(f"Realm {realm} set as default realm.", fg="green")


@realm_cli.command("clear_default")
def realm_clear_default():
    """
    Unset the default realm
    """
    set_default_realm(None)
    click.secho("cleared default realm.", fg="green")


config_cli.add_command(realm_cli)

resolver_cli = AppGroup("resolver", help="Manage user resolver")


@resolver_cli.command("list")
@click.option("-v", "--verbose", is_flag=True,
              help="Additionally output the configuration of the resolvers")
def resolver_list(verbose):
    """
    List the available resolvers and their type.
    """
    reso_list = get_resolver_list()

    if not verbose:
        for (name, resolver) in reso_list.items():
            click.echo(f"{name:16} - ({resolver.get('type')})")
    else:
        for (name, resolver) in reso_list.items():
            click.echo(f"{name:16} - ({resolver.get('type')})")
            click.echo("." * 32)
            data = resolver.get("data", {})
            for (k, v) in data.items():
                if k.lower() in ["bindpw", "password"]:
                    v = "xxxxx"
                click.echo(f"{k:>24}: {v!r}")
            click.echo("\n")


@resolver_cli.command("create")
@click.argument("name", type=str)
@click.argument("rtype", type=str)
@click.argument("conf_file", type=click.File())
def resolver_create(name, rtype, conf_file):
    """
    Create a new resolver with the specified name and type.

    The necessary resolver parameters are read from the file given with FILENAME.
    The file should contain a python dictionary.

    \b
    NAME:     The name of the resolver
    RTYPE:    The type of the resolver (can be ldapresolver, sqlresolver,
              httpresolver, passwdresolver or scimresolver)
    CONF_FILE: The name of the config file with the resolver parameters.

    \b
    Note: it might be more convenient to use the
        pi-manage config import
    command since it simplifies the import of the server configuration.
    """
    contents = conf_file.read()

    params = ast.literal_eval(contents)
    params["resolver"] = name
    params["type"] = rtype
    save_resolver(params)


@resolver_cli.command("create_internal")
@click.argument("name")
@click.pass_context
def resolver_create_internal(ctx, name):
    """
    This creates a new internal, editable sqlresolver. The users will be
    stored in the token database in a table called 'users_<NAME>'. You can then
    add this resolver to a new realm using the command
    'pi-manage config realm add <realm-name> <NAME>'.
    """
    sqluri = current_app.config.get("SQLALCHEMY_DATABASE_URI")
    sqlelements = sqluri.split("/")
    # mysql://user:password@localhost/pi
    # sqlite:////home/cornelius/src/privacyidea/data.sqlite
    sql_driver = sqlelements[0][:-1]
    user_pw_host = sqlelements[2]
    database = "/".join(sqlelements[3:])
    username = ""
    password = ""
    host = ""
    # determine host and user
    hostparts = user_pw_host.split("@")
    if len(hostparts) > 2:
        click.secho(f"Invalid database URI: {sqluri}", fg="red")
        ctx.exit(2)
    elif len(hostparts) == 1:
        host = hostparts[0] or "/"
    elif len(hostparts) == 2:
        host = hostparts[1] or "/"
        # split hostname and password
        userparts = hostparts[0].split(":")
        if len(userparts) == 2:
            username = userparts[0]
            password = userparts[1]
        elif len(userparts) == 1:
            username = userparts[0]
        else:
            click.secho(f"Invalid username and password in database URI: {sqluri}",
                        fg="red")
            ctx.exit(3)
    # now we can create the resolver
    params = {
        'resolver': name,
        'type': "sqlresolver",
        'Server': host,
        'Driver': sql_driver,
        'User': username,
        'Password': password,
        'Database': database,
        'Table': 'users_' + name,
        'Limit': '500',
        'Editable': '1',
        'Map': '{"userid": "id", "username": "username", '
               '"email":"email", "password": "password", '
               '"phone":"phone", "mobile":"mobile", "surname":"surname", '
               '"givenname":"givenname", "description": "description"}'}
    save_resolver(params)

    # Now we create the database table
    from sqlalchemy import create_engine
    from sqlalchemy import Table, MetaData, Column, Identity
    from sqlalchemy import Integer, String
    engine = create_engine(sqluri)
    metadata = MetaData()
    Table('users_%s' % name,
          metadata,
          Column('id', Integer, Identity(), primary_key=True),
          Column('username', String(40), unique=True),
          Column('email', String(80)),
          Column('password', String(255)),
          Column('phone', String(40)),
          Column('mobile', String(40)),
          Column('surname', String(40)),
          Column('givenname', String(40)),
          Column('description', String(255)))
    metadata.create_all(engine)


config_cli.add_command(resolver_cli)

event_cli = AppGroup("event", help="Manage events")


@event_cli.command("list")
def event_list():
    """
    List events
    """
    events = EventConfiguration().events
    click.echo("\n{0:7} {4:4} {1:30}\t{2:20}\t{3}".format("Active", "Name", "Module", "Action", "ID"))
    click.echo(90 * "=")
    for event in events:
        click.echo(f"{event['active']!r:7} {event['id']:<4} {event['name']:30}"
                   f"\t{event['handlermodule']:20}\t{event['action']}")


@event_cli.command("enable")
@click.argument("eid", type=int)
def event_enable(eid):
    """
    Enable event with id EID
    """
    try:
        r = enable_event(eid)
    except ResourceNotFoundError as e:
        click.secho(f"Could not enable event {eid}: {e!r}", fg="red")
    else:
        click.secho(f"Enabled Event with ID {r}", fg="green")


@event_cli.command("disable")
@click.argument("eid", type=int)
def event_disable(eid):
    """
    Disable event with id EID
    """
    try:
        r = enable_event(eid, enable=False)
    except ResourceNotFoundError as e:
        click.secho(f"Could not disable event {eid}: {e!r}", fg="red")
    else:
        click.secho(f"Disabled Event with ID {r}", fg="green")


@event_cli.command("delete")
@click.argument("eid", type=int)
def event_delete(eid):
    """
    Delete event with id EID
    """
    try:
        r = delete_event(eid)
    except ResourceNotFoundError as e:
        click.secho(f"Could not delete event with ID {eid}: {e!r}", fg="red")
    else:
        click.secho(f"Deleted Event with ID {r}", fg="green")


config_cli.add_command(event_cli)

policy_cli = AppGroup("policy", help="Manage policies")


@policy_cli.command("list")
def policy_list():
    """
    list the policies
    """
    pol_cls = PolicyClass()
    policies = pol_cls.list_policies()
    click.echo("Active   Name " + 16 * ' ' + "Scope")
    click.echo(40 * "=")
    for policy in policies:
        click.echo(f"{policy.get('active')!r:8} {policy.get('name'):20} "
                   f"{policy.get('scope')}")


@policy_cli.command("enable")
@click.argument("name", type=str)
def policy_enable(name):
    """
    enable the policy NAME.
    """
    try:
        enable_policy(name)
    except ResourceNotFoundError as e:
        click.secho(f"Could not enable policy {name}: {e!r}", fg="red")
    else:
        click.secho(f"Successfully enabled policy '{name}'", fg="green")


@policy_cli.command("disable")
@click.argument("name", type=str)
def policy_disable(name):
    """
    disable the policy NAME
    """
    try:
        enable_policy(name, enable=False)
    except ResourceNotFoundError as e:
        click.secho(f"Could not disable policy {name}: {e!r}", fg="red")
    else:
        click.secho(f"Successfully disabled policy '{name}'", fg="green")


@policy_cli.command("delete")
@click.argument("name", type=str)
def policy_delete(name):
    """
    delete the policy NAME
    """
    try:
        delete_policy(name)
    except ResourceNotFoundError as e:
        click.secho(f"Could not delete policy {name}: {e!r}", fg="red")
    else:
        click.secho(f"Successfully deleted policy '{name}'", fg="green")


@policy_cli.command("create")
@click.argument("name")
@click.argument("scope")
@click.argument("action")
@click.option("-f", "--file", type=click.File(),
              help="The file to import the policy configuration from.")
def policy_create(name, scope, action, file):
    """
    Create a new policy.

    The file given with the option '-f' must contain a Python dictionary and its
    content takes precedence over the other CLI arguments.
    I.e. if you are specifying a configuration file with '-f',
    the arguments name, scope and action need to be specified, but are ignored.

    \b
    Note: it might be more convenient to use the
        pi-manage config import
    command since it simplifies the import of the server configuration.
    """
    if file:
        try:
            contents = file.read()
            params = ast.literal_eval(contents)
            if params.get("name") and params.get("name") != name:
                print("Found name '{0!s}' in file, will use that instead of "
                      "'{1!s}'.".format(params.get("name"), name))
            else:
                print("name not defined in file, will use the cli value "
                      "{0!s}.".format(name))
                params["name"] = name

            if params.get("scope") and params.get("scope") != scope:
                print("Found scope '{0!s}' in file, will use that instead of "
                      "'{1!s}'.".format(params.get("scope"), scope))
            else:
                print("scope not defined in file, will use the cli value "
                      "{0!s}.".format(scope))
                params["scope"] = scope

            if params.get("action") and params.get("action") != action:
                print("Found action in file: '{0!s}', will use that instead "
                      "of: '{1!s}'.".format(params.get("action"), action))
            else:
                print("action not defined in file, will use the cli value "
                      "{0!s}.".format(action))
                params["action"] = action

            r = set_policy(params.get("name"),
                           scope=params.get("scope"),
                           action=params.get("action"),
                           realm=params.get("realm"),
                           resolver=params.get("resolver"),
                           user=params.get("user"),
                           time=params.get("time"),
                           client=params.get("client"),
                           active=params.get("active", True),
                           adminrealm=params.get("adminrealm"),
                           adminuser=params.get("adminuser"),
                           check_all_resolvers=params.get(
                               "check_all_resolvers", False))
            return r

        except Exception as _e:
            print("Unexpected error: {0!s}".format(sys.exc_info()[1]))

    else:
        r = set_policy(name, scope, action)
        return r


config_cli.add_command(policy_cli)

imp_fmt_dict = {
    'python': ast.literal_eval,
    'json': json.loads,
    'yaml': yaml.safe_load}


@config_cli.command("import")
@click.option('-i', '--input', "infile", type=click.File('r'),
              default=sys.stdin,
              help='The filename to import the data from. Read '
                   'from <stdin> if this argument is not given.')
@click.option('-t', '--types', multiple=True, default=['all'], show_default=True,
              type=click.Choice(['all'] + list(IMPORT_FUNCTIONS.keys()), case_sensitive=False),
              help='The types of configuration to import. By default import all '
                   'available data if a corresponding importer type exists. '
                   'Currently registered importer types are: '
                   '{0!s}'.format(', '.join(['all'] + list(IMPORT_FUNCTIONS.keys()))))
@click.option('-n', '--name', metavar="NAME",
              help='The name of the configuration object to import (default: import all)')
@click.pass_context
def config_import(ctx, infile, types, name):
    """
    Import server configuration using specific or all registered importer types.

    Note: Existing configuration with the same name will be overwritten, other
    existing configuration will be kept as is.
    """
    data = None
    imp_types = IMPORT_FUNCTIONS.keys() if 'all' in types else types

    content = infile.read()

    for fmt in imp_fmt_dict:
        try:
            data = imp_fmt_dict[fmt](content)
            break
        except (SyntaxError, json.decoder.JSONDecodeError, yaml.error.YAMLError) as _e:
            continue
    if not data:
        click.secho("Could not read input format! ", fg="red")
        click.secho(f"Accepted formats are: {', '.join(imp_fmt_dict.keys())}.",
                    fg="yellow")
        ctx.exit(1)

    def minver(s: str) -> str:
        """Get major.minor version number from string."""
        return ".".join(s.split('.')[0:2])

    # Check the version in the import data
    if "privacyIDEA_version" not in data:
        click.secho("Unable to determine version of exported data.", fg="yellow")
        click.secho("Please make sure that the imported configuration "
                    "works as expected.", fg="yellow")

    else:
        if minver(data["privacyIDEA_version"]) != minver(get_version_number()):
            click.secho(f"Version of export ({minver(data['privacyIDEA_version'])}) "
                        f"does not match current privacyIDEA "
                        f"version ({minver(get_version_number())}).", fg="yellow")
            click.secho("Please make sure that the imported configuration "
                        "works as expected.", fg="yellow")

    # we need to go through the importer functions based on priority
    for typ, value in sorted(IMPORT_FUNCTIONS.items(), key=lambda x: x[1]['prio']):
        if typ in imp_types:
            if typ in data:
                click.echo(f"Importing configuration type '{typ}'.")
                try:
                    value['func'](data[typ], name=name)
                except Exception as e:
                    click.secho(f"Could not successfully import data of "
                                f"type {typ}: {e!r}", fg="red")


# Create the "importer" command as a hidden and deprecated alias for "import"
importer_cmd = copy.copy(config_cli.get_command(None, "import"))
importer_cmd.hidden = True
importer_cmd.deprecated = True
importer_cmd.name = "importer"
importer_cmd.epilog = "This command is deprecated. Please use 'pi-manage config import' instead."
config_cli.add_command(importer_cmd)


exp_fmt_dict = {
    'python': str,
    'json': partial(json.dumps, indent=2),
    'yaml': yaml.safe_dump}


@config_cli.command("export")
@click.option('-o', '--output', type=click.File('w'),
              default=sys.stdout,
              help='The filename to export the data to. Write to '
                   '<stdout> if this argument is not given or is \'-\'.')
@click.option('-f', '--format', "fmt", default='python', show_default=True,
              type=click.Choice(exp_fmt_dict.keys(), case_sensitive=False),
              help='Output format, default is \'python\'')
# TODO: we need to have an eye on the help output, it might get less readable
#  when more exporter functions are added
@click.option('-t', '--types', multiple=True, default=['all'], show_default=True,
              type=click.Choice(['all'] + list(EXPORT_FUNCTIONS.keys()), case_sensitive=False),
              help='The types of configuration to export (can be given multiple '
                   'times). By default, export all available types. Currently '
                   'registered exporter types are: '
                   '{0!s}'.format(', '.join(['all'] + list(EXPORT_FUNCTIONS.keys()))),
              )
@click.option('-n', '--name', metavar="NAME",
              help='The name of the configuration object to export (default: export all)')
def config_export(output, fmt, types, name):
    """
    Export server configuration using specific or all registered exporter types.
    """
    exp_types = EXPORT_FUNCTIONS.keys() if 'all' in types else types

    out = {}
    for typ in exp_types:
        out.update({typ: EXPORT_FUNCTIONS[typ](name=name)})

    if out:
        # Add version information to output
        out["privacyIDEA_version"] = get_version_number()
        res = exp_fmt_dict.get(fmt.lower())(out) + '\n'
        output.write(res)


# Create the "exporter" command as a hidden and deprecated alias for "export"
exporter_cmd = copy.copy(config_cli.get_command(None, "export"))
exporter_cmd.hidden = True
exporter_cmd.deprecated = True
exporter_cmd.name = "exporter"
exporter_cmd.epilog = "This command is deprecated. Please use 'pi-manage config export' instead."
config_cli.add_command(exporter_cmd)

authcache_cli = AppGroup("authcache", help="Manage authentication cache")


@authcache_cli.command("cleanup")
@click.option("-m", "--minutes", default=480, show_default=True, type=int,
              help="Clean up authcache entries older than this number of minutes")
def authcache_cleanup(minutes):
    """
    Remove entries from the authcache.
    Remove all entries where the last_auth entry is older than the given number
    of minutes.
    """
    r = cleanup(minutes)
    click.echo(f"{r} entries deleted from authcache")


config_cli.add_command(authcache_cli)



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License: AGPLv3
#  contact: http://www.privacyidea.org
#
#  2022-02-03 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add verified enrollment
#  2018-06-06 Michael Becker <michael.becker@hs-niederrhein.de>
#             Add get_setting_type to make hotp.hashlib public and
#             therefore recognised in token enrollment with role user.
#  2018-01-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Set Yubikeys to be hardware tokenkind
#  2017-07-13 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add period to key uri for TOTP token
#
#  2016-04-29 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add get_default_settings to change the parameters before
#             the token is created
#  2014-10-03 Add getInitDetail
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This is the HOTP implementation.
It is inherited from lib.tokenclass and is thus dependent on models.py

This code is tested in tests/test_lib_tokens_hotp
"""

import time
import binascii

from .HMAC import HmacOtp
from privacyidea.api.lib.utils import getParam
from privacyidea.api.lib.policyhelper import get_init_tokenlabel_parameters
from privacyidea.lib.config import get_from_config
from privacyidea.lib.tokenclass import (TokenClass,
                                        TWOSTEP_DEFAULT_DIFFICULTY,
                                        TWOSTEP_DEFAULT_CLIENTSIZE,
                                        TOKENKIND)
from privacyidea.lib.log import log_with
from privacyidea.lib.apps import create_google_authenticator_url as cr_google
from privacyidea.lib.error import ParameterError
from privacyidea.lib.apps import create_oathtoken_url as cr_oath
from privacyidea.lib.utils import (create_img, is_true, b32encode_and_unicode,
                                   hexlify_and_unicode, determine_logged_in_userparams)
from privacyidea.lib.decorators import check_token_locked, check_token_otp_length
from privacyidea.lib.policy import SCOPE, ACTION, GROUP, Match
from privacyidea.lib.token import init_token
from privacyidea.lib.tokenclass import CLIENTMODE
from privacyidea.lib import _, lazy_gettext
import traceback
import logging

from passlib.crypto.digest import pbkdf2_hmac

from ..user import User

optional = True
required = False
log = logging.getLogger(__name__)

keylen = {'sha1': 20,
          'sha256': 32,
          'sha512': 64
          }

VERIFY_ENROLLMENT_MESSAGE = lazy_gettext("Please enter a valid OTP value of the new token.")


class HotpTokenClass(TokenClass):
    """
    hotp token class implementation
    """

    previous_otp_offset = 1

    # The HOTP token provides means to verify the enrollment
    can_verify_enrollment = True

    desc_hash_func = lazy_gettext('Specify the hashing function to be used. '
                                  'Can be SHA1, SHA256 or SHA512.')
    desc_otp_len = lazy_gettext('Specify the OTP length to be used. Can be 6 or 8 digits.')
    desc_key_gen = lazy_gettext("Force the key to be generated on the server.")
    desc_two_step_user = lazy_gettext('Specify whether users are allowed or forced to use '
                                      'two-step enrollment.')
    desc_two_step_admin = lazy_gettext('Specify whether admins are allowed or forced to '
                                       'use two-step enrollment.')

    @staticmethod
    def get_class_type():
        """
        return the token type shortname

        :return: 'hotp'
        :rtype: string
        """
        return "hotp"

    @staticmethod
    def get_class_prefix():
        """
        Return the prefix, that is used as a prefix for the serial numbers.
        :return: oath
        """
        return "OATH"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition
        Is used by lib.token.get_token_info

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict
        """
        res = {'type': 'hotp',
               'title': 'HOTP Event Token',
               'description': _('HOTP: Event based One Time Passwords.'),
               'user': ['enroll'],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of HOTP tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active HOTP tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       'hotp_2step_clientsize': {
                           'type': 'int',
                           'desc': _("The size of the OTP seed part contributed "
                                     "by the client (in bytes)")
                       },
                       'hotp_2step_serversize': {
                           'type': 'int',
                           'desc': _("The size of the OTP seed part contributed "
                                     "by the server (in bytes)")
                       },
                       'hotp_2step_difficulty': {
                           'type': 'int',
                           'desc': _("The difficulty factor used for the OTP "
                                     "seed generation (should be at least 10000)")
                       },
                       'hotp_' + ACTION.FORCE_APP_PIN: {
                           'type': 'bool',
                           'desc': _('Enforce setting an app pin for the privacyIDEA '
                                     'Authenticator App')
                       }
                   },
                   SCOPE.USER: {
                       'hotp_hashlib': {'type': 'str',
                                        'value': ["sha1",
                                                  "sha256",
                                                  "sha512"],
                                        'desc': HotpTokenClass.desc_hash_func},
                       'hotp_otplen': {'type': 'int',
                                       'value': [6, 8],
                                       'desc': HotpTokenClass.desc_otp_len},
                       'hotp_force_server_generate': {
                           'type': 'bool',
                           'desc': HotpTokenClass.desc_key_gen},
                       'hotp_2step': {'type': 'str',
                                      'value': ['allow', 'force'],
                                      'desc': HotpTokenClass.desc_two_step_user}
                   },
                   SCOPE.ADMIN: {
                       'hotp_hashlib': {'type': 'str',
                                        'value': ["sha1",
                                                  "sha256",
                                                  "sha512"],
                                        'desc': HotpTokenClass.desc_hash_func},
                       'hotp_otplen': {'type': 'int',
                                       'value': [6, 8],
                                       'desc': HotpTokenClass.desc_otp_len},
                       'hotp_2step': {'type': 'str',
                                      'value': ['allow', 'force'],
                                      'desc': HotpTokenClass.desc_two_step_admin}
                   }
               }
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res
        return ret

    @log_with(log)
    def __init__(self, db_token):
        """
        Create a new HOTP Token object

        :param db_token: instance of the orm db object
        :type db_token: DB object
        """
        TokenClass.__init__(self, db_token)
        self.set_type("hotp")
        self.hKeyRequired = True
        self.currently_in_challenge = False

    @log_with(log)
    def get_init_detail(self, params=None, user=None):
        """
        to complete the token initialization some additional details
        should be returned, which are displayed at the end of
        the token initialization.
        This is the e.g. the enrollment URL for a Google Authenticator.
        """
        response_detail = TokenClass.get_init_detail(self, params, user)
        params = params or {}
        tokenlabel = params.get("tokenlabel", "<s>")
        tokenissuer = params.get("tokenissuer", "privacyIDEA")
        # If the init_details contain an OTP key the OTP key
        # should be displayed as an enrollment URL
        otpkey = self.init_details.get('otpkey')
        # Add rollout state the response
        response_detail['rollout_state'] = self.token.rollout_state
        # Add two-step initialization parameters to response and QR code
        extra_data = {}
        if is_true(params.get("2stepinit")):
            twostep_parameters = self._get_twostep_parameters()
            extra_data.update(twostep_parameters)
            response_detail.update(twostep_parameters)
        imageurl = params.get("appimageurl")
        if imageurl:
            extra_data.update({"image": imageurl})
        force_app_pin = params.get(ACTION.FORCE_APP_PIN)
        if force_app_pin:
            extra_data.update({'pin': True})
        if otpkey:
            tok_type = self.type.lower()
            if user is not None:
                time_step = self.get_tokeninfo("timeStep", default="30")
                try:
                    key_bin = binascii.unhexlify(otpkey)
                    # also strip the padding =, as it will get problems with the google app.
                    value_b32_str = b32encode_and_unicode(key_bin).strip('=')
                    response_detail["otpkey"]["value_b32"] = value_b32_str
                    goo_url = cr_google(key=otpkey,
                                        user=user.login,
                                        realm=user.realm,
                                        tokentype=tok_type.lower(),
                                        serial=self.get_serial(),
                                        tokenlabel=tokenlabel,
                                        hash_algo=self.hashlib,
                                        digits=params.get("otplen", 6),
                                        period=time_step,
                                        issuer=tokenissuer,
                                        user_obj=user,
                                        extra_data=extra_data)
                    response_detail["googleurl"] = {"description": _("URL for google Authenticator"),
                                                    "value": goo_url,
                                                    "img": create_img(goo_url)
                                                    }

                    oath_url = cr_oath(otpkey=otpkey,
                                       user=user.login,
                                       realm=user.realm,
                                       type=tok_type,
                                       serial=self.get_serial(),
                                       tokenlabel=tokenlabel,
                                       extra_data=extra_data)
                    response_detail["oathurl"] = {"description": _("URL for"
                                                                   " OATH "
                                                                   "token"),
                                                  "value": oath_url,
                                                  "img": create_img(oath_url)
                                                  }
                except KeyError as ex:
                    log.debug("{0!s}".format((traceback.format_exc())))
                    log.error('Unknown Tag {0!s} in one of your policy definition'
                              .format(ex))
                except Exception as ex:  # pragma: no cover
                    log.debug("{0!s}".format((traceback.format_exc())))
                    log.error('failed to set oath or google url: {0!r}'.format(ex))

        return response_detail

    def _get_twostep_parameters(self):
        """
        :return: A dictionary with the keys ``2step_salt``,
        ``2step_difficulty``, ``2step_output``, mapping each key to an integer.
        """
        return {'2step_salt': int(self.get_tokeninfo('2step_clientsize')),
                '2step_output': int(keylen[self.hashlib]),
                '2step_difficulty': int(self.get_tokeninfo('2step_difficulty'))}

    @log_with(log)
    def update(self, param, reset_failcount=True):
        """
        process the initialization parameters

        Do we really always need an otpkey?
        the otpKey is handled in the parent class

        :param param: dict of initialization parameters
        :type param: dict
        :param reset_failcount: reset the failcount
        :type reset_failcount: bool
        """
        # In case am Immutable MultiDict:
        upd_param = {}
        for k, v in param.items():
            upd_param[k] = v

        # Special handling of 2-step enrollment
        if is_true(getParam(param, "2stepinit", optional)):
            # Use the 2step_serversize setting for the size of the server secret
            # (if it is set)
            if "2step_serversize" in upd_param:
                upd_param["keysize"] = int(getParam(upd_param, "2step_serversize", required))
            # Add twostep settings to the tokeninfo
            for key, default in [
                ("2step_difficulty", TWOSTEP_DEFAULT_DIFFICULTY),
                ("2step_clientsize", TWOSTEP_DEFAULT_CLIENTSIZE)]:
                self.add_tokeninfo(key, getParam(param, key, optional, default))

        val = getParam(upd_param, "hashlib", optional)
        if val is not None:
            hashlibStr = val
        else:
            hashlibStr = self.hashlib

        # check if the key_size is provided
        # if not, we could derive it from the hashlib
        key_size = getParam(upd_param, 'key_size', optional) \
                   or getParam(upd_param, 'keysize', optional)
        if key_size is None:
            upd_param['keysize'] = keylen.get(hashlibStr)

        otpKey = getParam(upd_param, "otpkey", optional)
        genkey = is_true(getParam(upd_param, "genkey", optional))
        if genkey and otpKey:
            # The Base TokenClass does not allow otpkey and genkey at the
            # same time
            del upd_param['otpkey']
        upd_param['hashlib'] = hashlibStr
        # We first need to call the parent class. Since exceptions would be
        # raised here.
        TokenClass.update(self, upd_param, reset_failcount)

        self.add_tokeninfo("hashlib", hashlibStr)

        # check the tokenkind
        if self.token.serial.startswith("UB"):
            self.add_tokeninfo("tokenkind", TOKENKIND.HARDWARE)

    @property
    def hashlib(self):
        hashlibStr = self.get_tokeninfo("hashlib") or \
                     get_from_config("hotp.hashlib", 'sha1')
        return hashlibStr

    def _calc_otp(self, counter):
        """
        Helper function to calculate the OTP value for the given counter

        :param counter: The OTP counter
        :return: OTP value as string
        """
        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()
        hmac2Otp = HmacOtp(secretHOtp,
                           self.get_otp_count(),
                           otplen,
                           self.get_hashlib(self.hashlib))

        otpval = hmac2Otp.generate(counter=counter,
                                   inc_counter=False,
                                   do_truncation=True)
        return otpval

    @log_with(log)
    @check_token_otp_length
    @check_token_locked
    def check_otp(self, anOtpVal, counter=None, window=None, options=None):
        """
        check if the given OTP value is valid for this token.

        :param anOtpVal: the to be verified otpvalue
        :type anOtpVal: string
        :param counter: the counter state, that should be verified
        :type counter: int
        :param window: the counter +window, which should be checked
        :type window: int
        :param options: the dict, which could contain token specific info
        :type options: dict
        :return: the counter state or -1
        :rtype: int
        """
        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()

        if counter is None:
            counter = int(self.get_otp_count())
        if window is None:
            window = int(self.get_count_window())
        hmac2Otp = HmacOtp(secretHOtp,
                           counter,
                           otplen,
                           self.get_hashlib(self.hashlib))
        res = hmac2Otp.checkOtp(anOtpVal, window)

        if res == -1:
            res = self._autosync(hmac2Otp, anOtpVal)
        if res != -1:
            # on success, we save the counter
            self.set_otp_count(res + 1)
            # We could also store it temporarily
            # self.auth_details["matched_otp_counter"] = res

        return res

    @log_with(log)
    def check_otp_exist(self, otp: str, window: int = 10, symetric: bool = False, inc_counter: bool = True,
                        counter: int = None) -> int:
        """
        checks if the given OTP value is/are values of this very token.
        This is used to autoassign and to determine the serial number of
        a token.

        :param otp: the to be verified otp value
        :param window: the lookahead window for the counter
        :param inc_counter: True if the counter shall be increased for a valid OTP, False otherwise
        :param counter: The counter to be used to calculate the correct OTP value. If None the internal counter of the
            token is used.
        :return: counter or -1 if otp does not exist
        """
        res = -1
        otplen = int(self.token.otplen)
        counter = counter or int(self.token.count)

        secretHOtp = self.token.get_otpkey()
        hmac2Otp = HmacOtp(secretHOtp, counter, otplen,
                           self.get_hashlib(self.hashlib))
        res = hmac2Otp.checkOtp(otp, window, symetric=symetric)

        if inc_counter and res >= 0:
            # As usually the counter is increased in lib.token.checkUserPass,
            # we need to do this manually here:
            self.inc_otp_counter(res)
        if res == -1:
            msg = "otp counter {0!r} was not found".format(otp)
        else:
            msg = "otp counter {0!r} was found".format(otp)
        log.debug("end. {0!r}: res {1!r}".format(msg, res))
        return res

    @log_with(log)
    def is_previous_otp(self, otp):
        """
        Check if the OTP values was previously used.

        :param otp:
        :return:
        """
        counter = int(self.get_otp_count())
        if counter > 0:
            previous_otp = self._calc_otp(counter - self.previous_otp_offset)
            res = previous_otp == otp
            if res:
                log.info("Previous OTP used again. "
                         "Serial {0!s} with counter {1!s}.".format(self.token.serial, counter))
            return res
        else:
            # The internal counter is 0, the token was not used, yet.
            return False

    @log_with(log)
    def _autosync(self, hmac2Otp, anOtpVal):
        """
        automatically sync the token based on two otp values
        internal method to implement the _autosync within the
        checkOtp method.

        :param hmac2Otp: the hmac object (with reference to the token secret)
        :type hmac2Otp: hmac object

        :param anOtpVal: the actual otp value
        :type anOtpVal: string

        :return: counter or -1 if otp does not exist
        :rtype:  int
        """
        res = -1
        # get _autosync from config or use False as default
        autosync = get_from_config("AutoResync", False, return_bool=True)

        # if _autosync is not enabled
        if autosync is False:
            log.debug("end. _autosync is not enabled : res {0!r}".format(res))
            return res

        info = self.get_tokeninfo()
        syncWindow = self.get_sync_window()

        # check if the otpval is valid in the sync scope
        res = hmac2Otp.checkOtp(anOtpVal, syncWindow)

        # If the otpval is valid in the big sync scope, we
        # either store the value in the tokeninfo
        # or see if already another value exists.
        if res != -1:
            # if former is defined
            if "otp1c" in info:
                # check if this is consecutive
                otp1c = int(info.get("otp1c"))
                otp2c = res

                if (otp1c + 1) != otp2c:
                    res = -1

                if "dueDate" in info:
                    dueDate = int(info.get("dueDate"))
                    now = int(time.time())
                    if dueDate <= now:
                        res = -1
                else:
                    # if by any reason the dueDate is missing!
                    res = -1  # pragma: no cover

                # now clean the resync data
                self.del_tokeninfo("dueDate")
                self.del_tokeninfo("otp1c")

            else:
                self.add_tokeninfo("otp1c", res)
                self.add_tokeninfo("dueDate", int(time.time()) +
                                   self.get_sync_timeout())

                res = -1

        return res

    @log_with(log)
    def resync(self, otp1, otp2, options=None):
        """
        resync the token based on two otp values

        :param otp1: the first otp value
        :type otp1: string

        :param otp2: the second otp value
        :type otp2: string

        :param options: optional token specific parameters
        :type options:  dict or None

        :return: counter or -1 if otp does not exist
        :rtype:  int
        """
        ret = False
        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()
        counter = self.token.count
        syncWindow = self.get_sync_window()
        # log.debug("serial: %s",serialNum)
        hmac2Otp = HmacOtp(secretHOtp, counter, otplen,
                           self.get_hashlib(self.hashlib))
        counter = hmac2Otp.checkOtp(otp1, syncWindow)

        if counter == -1:
            log.debug("exit. First counter (-1) not found  ret: {0!r}".format(ret))
            return ret

        nextOtp = hmac2Otp.generate(counter + 1)

        if nextOtp != otp2:
            log.debug("exit. Failed to verify second otp: nextOtp: "
                      "%r != otp2: %r ret: %r" % (nextOtp, otp2, ret))
            return ret

        ret = True
        self.inc_otp_counter(counter + 1, reset=True)

        log.debug("end. resync was successful: ret: {0!r}".format(ret))
        return ret

    @staticmethod
    def get_sync_timeout():
        """
        get the token sync timeout value

        :return: timeout value in seconds
        :rtype:  int
        """
        try:
            timeOut = int(get_from_config("AutoResyncTimeout", 5 * 60))
        except Exception as ex:
            log.warning("AutoResyncTimeout: value error {0!r} - reset to 5*60".format(ex))
            timeOut = 5 * 60

        return timeOut

    @log_with(log)
    def get_otp(self, current_time=None):
        """
        return the next otp value

        :param current_time: Not Used in HOTP
        :return: next otp value and PIN if possible
        :rtype: tuple
        """
        otplen = int(self.token.otplen)
        secretHOtp = self.token.get_otpkey()

        hmac2Otp = HmacOtp(secretHOtp,
                           self.token.count,
                           otplen,
                           self.get_hashlib(self.hashlib))
        otpval = hmac2Otp.generate(inc_counter=False)

        pin = self.token.get_pin()

        if get_from_config("PrependPin") == "True":
            combined = "{0!s}{1!s}".format(pin, otpval)
        else:
            combined = "{0!s}{1!s}".format(otpval, pin)

        return 1, pin, otpval, combined

    @log_with(log)
    def get_multi_otp(self, count=0, epoch_start=0, epoch_end=0,
                      curTime=None, timestamp=None,
                      counter_index=False):
        """
        return a dictionary of multiple future OTP values of the
        HOTP/HMAC token

        WARNING: the dict that is returned contains a sequence number as key.
                 This it NOT the otp counter!

        :param count: how many otp values should be returned
        :type count: int
        :param epoch_start: Not used in HOTP
        :param epoch_end: Not used in HOTP
        :param curTime: Not used in HOTP
        :param timestamp: not used in HOTP
        :param counter_index: whether the counter should be used as index
        :return: tuple of status: boolean, error: text and the OTP dictionary
        """
        otp_dict = {"type": "hotp", "otp": {}}
        ret = False
        error = "No count specified"
        otplen = int(self.token.otplen)

        secretHOtp = self.token.get_otpkey()
        hmac2Otp = HmacOtp(secretHOtp, self.token.count, otplen,
                           self.get_hashlib(self.hashlib))
        log.debug("retrieving {0:d} OTP values for token {1!s}".format(count, hmac2Otp))

        if count > 0:
            error = "OK"
            for i in range(count):
                otpval = hmac2Otp.generate(self.token.count + i,
                                           inc_counter=False)
                if counter_index:
                    otp_dict["otp"][self.token.count + i] = otpval
                else:
                    otp_dict["otp"][i] = otpval
            ret = True

        return ret, error, otp_dict

    @staticmethod
    def get_setting_type(key):
        settings = {"hotp.hashlib": "public"}
        return settings.get(key, "")

    @classmethod
    def get_default_settings(cls, g, params):
        """
        This method returns a dictionary with default settings for token
        enrollment.
        These default settings are defined in SCOPE.USER or SCOPE.ADMIN and are
        hotp_hashlib, hotp_otplen.
        If these are set, the user or admin will only be able to enroll tokens
        with these values.

        The returned dictionary is added to the parameters of the API call.

        :param g: context object, see documentation of ``Match``
        :param params: The call parameters
        :type params: dict
        :return: default parameters
        """
        ret = {}
        if not g.logged_in_user:
            return ret
        (role, username, userrealm, adminuser, adminrealm) = determine_logged_in_userparams(g.logged_in_user,
                                                                                            params)
        return cls._get_default_settings(g, role, username, userrealm, adminuser, adminrealm)

    @classmethod
    def _get_default_settings(cls, g, role="user", username=None, userrealm=None,
                              adminuser=None, adminrealm=None):
        """
        Internal function that can be called either during enrollment via /token/init or during
        enrollment via validate/check.
        This way we have consistent policy handling.
        """
        ret = {}
        hashlib_pol = Match.generic(g, scope=role,
                                    action="hotp_hashlib",
                                    user=username,
                                    realm=userrealm,
                                    adminrealm=adminrealm,
                                    adminuser=adminuser).action_values(unique=True)
        if hashlib_pol:
            ret["hashlib"] = list(hashlib_pol)[0]

        otplen_pol = Match.generic(g, scope=role,
                                   action="hotp_otplen",
                                   user=username,
                                   realm=userrealm,
                                   adminrealm=adminrealm,
                                   adminuser=adminuser).action_values(unique=True)
        if otplen_pol:
            ret["otplen"] = list(otplen_pol)[0]
        return ret

    def generate_symmetric_key(self, server_component, client_component,
                               options=None):
        """
        Generate a composite key from a server and client component
        using a PBKDF2-based scheme.

        :param server_component: The component usually generated by privacyIDEA
        :type server_component: hex string
        :param client_component: The component usually generated by the
            client (e.g. smartphone)
        :type client_component: hex string
        :param options:
        :return: the new generated key as hex string
        :rtype: str
        """
        # As /token/init has already been called before, self.hashlib
        # is already set.
        keysize = keylen[self.hashlib]
        rounds = int(self.get_tokeninfo('2step_difficulty'))
        decoded_client_component = binascii.unhexlify(client_component)
        expected_client_size = int(self.get_tokeninfo('2step_clientsize'))
        if expected_client_size != len(decoded_client_component):
            raise ParameterError('Client Secret Size is expected to be {}, but is {}'.format(
                expected_client_size, len(decoded_client_component)
            ))
        # Based on the two components, we generate a symmetric key using PBKDF2
        # We pass the hex-encoded server component as the password and the
        # client component as the salt.
        secret = pbkdf2_hmac(digest='sha1',
                             secret=server_component.lower(),
                             salt=decoded_client_component,
                             rounds=rounds,
                             keylen=keysize)
        return hexlify_and_unicode(secret)

    @staticmethod
    def get_import_csv(l):
        """
        Read the list from a csv file and return a dictionary, that can be used
        to do a token_init.

        :param l: The list of the line of a csv file
        :type l: list
        :return: A dictionary of init params
        """
        params = TokenClass.get_import_csv(l)
        # get Counter
        if len(l) >= 5:
            params["counter"] = int(l[4].strip())
        return params

    def prepare_verify_enrollment(self, options=None):
        """
        This is called, if the token should be enrolled in a way, that the user
        needs to provide a proof, that the server can verify, that the token
        was successfully enrolled. E.g. with HOTP tokens the user might need to provide
        a correct OTP value.

        The returned dictionary is added to the response in "detail" -> "verify".

        :return: A dictionary with information that is needed to trigger the verification.
        """
        return {"message": str(VERIFY_ENROLLMENT_MESSAGE)}

    def verify_enrollment(self, verify):
        """
        This is called during the 2nd step of the verified enrollment.
        This method verifies the actual response from the user.
        Returns true, if the verification was successful.

        :param verify: The response given by the user
        :return: True
        """
        r = self.check_otp(verify)
        log.debug("Enrollment verified: {0!s}".format(r))
        return r >= 0

    @classmethod
    def enroll_via_validate(cls, g, content, user_obj, message=None):
        """
        This class method is used in the policy ENROLL_VIA_MULTICHALLENGE.
        It enrolls a new token of this type and returns the necessary information
        to the client by modifying the content.

        :param g: context object
        :param content: The content of a response
        :param user_obj: A user object
        :param message: An alternative message displayed to the user during enrollment
        :return: None, the content is modified
        """
        message = message or _("Please scan the QR code and enter the OTP value!")

        # Get the user policy token settings like otplen and hashlib
        params = cls._get_default_settings(g, username=user_obj.login, userrealm=user_obj.realm)
        params["type"] = cls.get_class_type()
        params["genkey"] = 1
        token_obj = init_token(params, user=user_obj)
        content.get("result")["value"] = False
        content.get("result")["authentication"] = "CHALLENGE"

        detail = content.setdefault("detail", {})
        # Create a challenge!
        c = token_obj.create_challenge()
        # get details of token
        enroll_params = get_init_tokenlabel_parameters(g, user_object=user_obj,
                                                       token_type=cls.get_class_type())
        if "hashlib" not in enroll_params:
            enroll_params["hashlib"] = token_obj.hashlib
        if "otplen" not in enroll_params:
            enroll_params["otplen"] = token_obj.token.otplen
        if "timeStep" not in enroll_params and params.get("type").lower() == "totp":
            enroll_params["timeStep"] = token_obj.get_tokeninfo("timeStep", 30)

        init_details = token_obj.get_init_detail(params=enroll_params,
                                                 user=user_obj)
        detail["transaction_ids"] = [c[2]]
        detail["messages"] = [message]
        chal = {"transaction_id": c[2],
                "image": init_details.get("googleurl").get("img"),
                "link": init_details.get("googleurl").get("value"),
                "client_mode": CLIENTMODE.INTERACTIVE,
                "serial": token_obj.token.serial,
                "type": token_obj.type,
                "message": message}
        detail["multi_challenge"] = [chal]
        detail.update(chal)

    def has_further_challenge(self, options=None):
        """

        :param options:
        :return:
        """
        try:
            return self.currently_in_challenge
        except AttributeError:
            # Certain from HOTP inherited tokenclasses might not set currently_in_challenge
            return False

    @classmethod
    def is_multichallenge_enrollable(cls):
        return True

    def get_enroll_url(self, user: User, params: dict) -> str:
        """
        Return the URL to enroll this token.

        :param user: The user object
        :param params: Further parameters
        :return: The URL containing all required information to enroll the token
        """
        token_secret = self.token.get_otpkey().getKey().decode("utf-8")
        self.init_details.update({"otpkey": token_secret})
        init_details = self.get_init_detail(params, user)
        enroll_url = init_details.get("googleurl").get("value")
        return enroll_url



#  2016-12-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2016. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module for calling external scripts.
You can put any script in the directory specified in
PI_SCRIPT_HANDLER_DIRECTORY and call this script in any event.

The scripts can take parameters like
* serial
* user
* realm
* resolver
* logged_in_user
"""
from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib.utils import is_true
from privacyidea.lib.framework import get_app_config_value
from privacyidea.lib.error import ServerError, ConfigAdminError
from privacyidea.lib import _
from privacyidea.app import db
import logging
import subprocess  # nosec B404 # We know what we are doing and only allow trusted script calls
import os
import traceback

log = logging.getLogger(__name__)

SCRIPT_BACKGROUND = "background"
SCRIPT_WAIT = "wait"


class ScriptEventHandler(BaseEventHandler):
    """
    An Eventhandler needs to return a list of actions, which it can handle.
    It also returns a list of allowed action and conditions
    It returns an identifier, which can be used in the eventhandling definitions
    """

    identifier = "Script"
    description = "This event handler can trigger external scripts."

    def __init__(self):
        super(ScriptEventHandler, self).__init__()
        self.script_directory = get_app_config_value("PI_SCRIPT_HANDLER_DIRECTORY",
                                                     "/etc/privacyidea/scripts")

    @property
    def allowed_positions(self):
        """
        This returns the allowed positions of the event handler definition.

        :return: list of allowed positions
        """
        return ["post", "pre"]

    @property
    def actions(self):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        try:
            scripts = os.listdir(self.script_directory)
        except (FileNotFoundError, PermissionError) as e:
            log.error(f"Unable to read script directory: {e}")
            raise ConfigAdminError(f"Unable to open script directory: {e}")
        actions = {}
        for script in scripts:
            actions[script] = {
                "background": {
                    "type": "str",
                    "required": True,
                    "description": _("Wait for script to complete or run script in background. This will "
                                     "either return the HTTP request early or could also block the request."),
                    "value": [SCRIPT_BACKGROUND, SCRIPT_WAIT]
                },
                "raise_error": {
                    "type": "bool",
                    "visibleIf": "background",
                    "visibleValue": SCRIPT_WAIT,
                    "description": _("On script error raise exception in HTTP request.")
                },
                "sync_to_database": {
                    "type": "bool",
                    "description": _("Finish current transaction before running "
                                     "the script. This is useful if changes to "
                                     "the database should be made available to "
                                     "the script or the running request.")
                },
                "serial": {
                    "type": "bool",
                    "description": _("Add '--serial <serial number>' as script "
                                     "parameter.")
                },
                "user": {
                    "type": "bool",
                    "description": _("Add '--user <username>' as script "
                                     "parameter.")
                },
                "realm": {
                    "type": "bool",
                    "description": _("Add '--realm <realm>' as script "
                                     "parameter.")
                },
                "logged_in_user": {
                    "type": "bool",
                    "description": _("Add the username of the logged in user "
                                     "as script parameter like "
                                     "'--logged_in_user <username>'.")
                },
                "logged_in_role": {
                    "type": "bool",
                    "description": _("Add the role (either admin or user) of "
                                     "the logged in user as script parameter "
                                     "like '--logged_in_role <role>'.")
                }
            }

        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        """
        ret = True
        script_name = self.script_directory + "/" + action
        proc_args = [script_name]

        g = options.get("g")
        request = options.get("request")
        response = options.get("response")
        content = self._get_response_content(response)
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})

        if hasattr(g, "logged_in_user"):
            logged_in_user = g.logged_in_user
        else:
            logged_in_user = {"username": "none",
                              "realm": "none",
                              "role": "none"}

        serial = (request.all_data.get("serial")
                  or content.get("detail", {}).get("serial")
                  or g.audit_object.audit_data.get("serial"))

        if is_true(handler_options.get("serial")):
            proc_args.append("--serial")
            proc_args.append(serial or "none")

        if is_true(handler_options.get("user")):
            proc_args.append("--user")
            proc_args.append(request.User.login or "none")

        if is_true(handler_options.get("realm")):
            proc_args.append("--realm")
            proc_args.append(request.User.realm or "none")

        if is_true(handler_options.get("logged_in_user")):
            proc_args.append("--logged_in_user")
            proc_args.append("{username}@{realm}".format(
                **logged_in_user))

        if is_true(handler_options.get("logged_in_role")):
            proc_args.append("--logged_in_role")
            proc_args.append(logged_in_user.get("role", "none"))

        rcode = 0
        try:
            log.info("Starting script {script!r}.".format(script=script_name))
            if is_true(handler_options.get('sync_to_database', False)):
                log.debug('Committing current transaction for script '
                          '{0!s}'.format(script_name))
                db.session.commit()
            # Trusted input/no user input: The scripts are created by user root and read from hard disk
            p = subprocess.Popen(proc_args, cwd=self.script_directory, universal_newlines=True)  # nosec B603
            if handler_options.get("background") == SCRIPT_WAIT:
                rcode = p.wait()

        except Exception as e:
            log.warning("Failed to execute script {0!r}: {1!r}".format(
                script_name, e))
            log.warning(traceback.format_exc())
            if handler_options.get("background") == SCRIPT_WAIT and is_true(handler_options.get("raise_error")):
                raise ServerError("Failed to start script.")

        if rcode:
            log.warning("Script {script!r} failed to execute with error code {error!r}".format(script=script_name,
                                                                                               error=rcode))
            if is_true(handler_options.get("raise_error")):
                raise ServerError("Error during execution of the script.")

        return ret



# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import datetime
import re
import sys
import click
import yaml
from flask import current_app
from flask.cli import AppGroup
from sqlalchemy import create_engine, desc, MetaData
from sqlalchemy.orm import sessionmaker

from privacyidea.lib.audit import getAudit
from privacyidea.lib.auditmodules.sqlaudit import LogEntry
from privacyidea.lib.sqlutils import delete_matching_rows
from privacyidea.lib.utils import parse_timedelta

audit_cli = AppGroup("audit", help="Manage Audit log")


@audit_cli.command("rotate")
@click.option('-hw', '--highwatermark', default=10000, show_default=True,
              help="If entries exceed this value, old entries are deleted.")
@click.option('-lw', '--lowwatermark', default=5000, show_default=True,
              help="Keep this number of entries.")
@click.option('--age', default=0,
              help="Delete audit entries older than these number of days.")
@click.option('--config', type=click.File('r'),
              help="Read config from the specified yaml file.")
@click.option('--dryrun', is_flag=True,
              help="Do not actually delete, only show what would be done.")
@click.option('--chunksize', type=int,
              help="Delete entries in chunks of the given size to avoid deadlocks")
def rotate_audit(highwatermark, lowwatermark, age, config,
                 dryrun, chunksize):
    """
    Clean the SQL audit log.

    You can either clean the audit log based on the number of entries of
    based on the age of the entries.

    Cleaning based on number of entries:

    If more than 'highwatermark' entries are in the audit log old entries
    will be deleted, so that 'lowwatermark' entries remain.

    Cleaning based on age:

    Entries older than the specified number of days are deleted.

    Cleaning based on config file:

    You can clean different type of entries with different ages or watermark.
    See the documentation for the format of the config file
    """
    metadata = MetaData()
    if chunksize is not None:
        chunksize = int(chunksize)

    default_module = "privacyidea.lib.auditmodules.sqlaudit"
    token_db_uri = current_app.config.get("SQLALCHEMY_DATABASE_URI")
    audit_db_uri = current_app.config.get("PI_AUDIT_SQL_URI", token_db_uri)
    audit_module = current_app.config.get("PI_AUDIT_MODULE", default_module)
    if audit_module != default_module:
        click.secho(f"We only rotate SQL audit module. You are using {audit_module}",
                    fg="red")
    if config:
        click.echo("Cleaning up with config file.")
    elif age > 0:
        click.echo(f"Cleaning up with age: {age}.")
    else:
        click.echo(f"Cleaning up with high: {highwatermark}, low: {lowwatermark}.")

    engine = create_engine(audit_db_uri)
    # create a configured "Session" class
    session = sessionmaker(bind=engine)()
    # create a Session
    metadata.create_all(engine)
    if config:
        yml_config = yaml.safe_load(config)
        auditlogs = session.query(LogEntry).all()
        delete_list = []
        for log in auditlogs:
            click.echo("investigating log entry {0!s}".format(log.id))
            for rule in yml_config:
                age = int(rule.get("rotate"))
                rotate_date = datetime.datetime.now() - datetime.timedelta(days=age)

                match = False
                for key in rule.keys():
                    if key not in ["rotate"]:
                        search_value = str(rule.get(key))
                        click.echo(" + searching for {0!r} in {1!s}".format(search_value,
                                                                            getattr(LogEntry, key)))
                        audit_value = str(getattr(log, key, ""))
                        m = re.search(search_value, audit_value)
                        if m:
                            # it matches!
                            click.echo(" + -- found {0!r}".format(audit_value))
                            match = True
                        else:
                            # It does not match, we continue to next rule
                            click.echo(" + NO MATCH - SKIPPING rest of conditions!")
                            match = False
                            break

                if match:
                    if log.date < rotate_date:
                        # Delete it!
                        click.echo(" + Deleting {0!s} due to rule {1!s}".format(log.id, rule))
                        # Delete it
                        delete_list.append(log.id)
                    # skip all other rules and go to the next log entry
                    break
        if dryrun:
            click.echo("If you only would let me I would clean up "
                       "{0!s} entries!".format(len(delete_list)))
        else:
            click.echo("Cleaning up {0!s} entries.".format(len(delete_list)))
            delete_matching_rows(session, LogEntry.__table__,
                                 LogEntry.id.in_(delete_list), chunksize)
    elif age:
        now = datetime.datetime.now() - datetime.timedelta(days=age)
        click.echo("Deleting entries older than {0!s}".format(now))
        criterion = LogEntry.date < now
        if dryrun:
            r = LogEntry.query.filter(criterion).count()
            click.echo("Would delete {0!s} entries.".format(r))
        else:
            r = delete_matching_rows(session, LogEntry.__table__, criterion, chunksize)
            click.echo("{0!s} entries deleted.".format(r))
    else:
        count = session.query(LogEntry.id).count()
        last_id = 0
        for l in session.query(LogEntry.id).order_by(desc(LogEntry.id)).limit(1):
            last_id = l[0]
        click.echo("The log audit log has %i entries, the last one is %i" % (count,
                                                                             last_id))
        # deleting old entries
        if count > highwatermark:
            click.echo("More than %i entries, deleting..." % highwatermark)
            cut_id = last_id - lowwatermark
            # delete all entries less than cut_id
            click.echo("Deleting entries smaller than %i" % cut_id)
            criterion = LogEntry.id < cut_id
            if dryrun:
                r = LogEntry.query.filter(criterion).count()
                click.echo("Would delete {0!s} entries.".format(r))
            else:
                r = delete_matching_rows(session, LogEntry.__table__, criterion, chunksize)
                click.echo("{0!s} entries deleted.".format(r))


def _validate_timelimit(_ctx, _param, value):
    if value:
        try:
            tl = parse_timedelta(value)
            return tl
        except TypeError as e:
            raise click.BadParameter(f"Could not parse timelimit ({value}): {e}")
    return None


@audit_cli.command("dump")
@click.option('-t', '--timelimit', callback=_validate_timelimit,
              help="Limit the dumped audit entries to a certain period "
                   "(i.e. '5d' or '3h' for the entries from the last five days "
                   "or three hours. By default all audit entries will be dumped.")
@click.option('-f', '--filename', type=click.File('w'), default=sys.stdout,
              help="Name of the file to dump the audit entries into. "
                   "By default write to stdout.")
def dump_audit(filename, timelimit):
    """Dump the audit log in csv format."""
    audit = getAudit(current_app.config)
    for line in audit.csv_generator(timelimit=timelimit):
        filename.write(line)



"""
This test file tests the lib.tokens.vascotoken
This depends on lib.tokenclass
"""
import functools
from binascii import hexlify

import mock

from privacyidea.lib.error import ParameterError
from privacyidea.lib.token import check_serial_pass
from privacyidea.lib.tokens.vascotoken import VascoTokenClass
from privacyidea.models import Token
from tests.base import MyTestCase


def mock_verification(replacement):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            with mock.patch('privacyidea.lib.tokens.vasco.vasco_dll') as mock_dll:
                mock_dll.AAL2VerifyPassword.side_effect = replacement
                return f(*args, **kwargs)
        return wrapper
    return decorator

def _set_next_blob(data):
    """
    update ``data._obj.Blob``: Increment the contents by 1.
    i.e. "XXX...X" is replaced with "YYY...Y".
    """
    data._obj.Blob = b"".join(bytes((x + 1,)) for x in data._obj.Blob)

def mock_success(data, params, password, challenge):
    # fake a new blob
    _set_next_blob(data)
    return 0

def create_mock_failure(return_value):
    def mock_failure(data, params, password, challenge):
        # fake a new blob
        _set_next_blob(data)
        return return_value
    return mock_failure

def mock_missing_dll(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        with mock.patch('privacyidea.lib.tokens.vasco.vasco_dll', None):
            return f(*args, **kwargs)
    return wrapper

class VascoTokenTest(MyTestCase):
    otppin = "topsecret"
    serial1 = "ser1"
    serial2 = "ser2"
    resolvername1 = "resolver1"
    resolvername2 = "Resolver2"
    resolvername3 = "reso3"
    realm1 = "realm1"
    realm2 = "realm2"

    def test_01_create_token(self):
        db_token = Token(self.serial1, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X" * 248).decode("utf-8"),
                      "pin": self.otppin})
        self.assertTrue(token.token.serial == self.serial1, token)
        self.assertTrue(token.token.tokentype == "vasco", token.token.tokentype)
        self.assertTrue(token.type == "vasco", token)
        class_prefix = token.get_class_prefix()
        self.assertTrue(class_prefix == "VASC", class_prefix)
        self.assertTrue(token.get_class_type() == "vasco", token)

    def test_02_genkey_fails(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        self.assertRaises(ParameterError, token.update, {"genkey": "1"})
        token.delete_token()

    @mock_missing_dll
    def test_03_no_vasco_library(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X"*248).decode("utf-8"),
                      "pin": self.otppin})
        self.assertRaises(RuntimeError, token.authenticate, "{}123456".format(self.otppin))
        token.delete_token()

    @mock_verification(create_mock_failure(1))
    def test_04_failure(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X"*248).decode("utf-8"),
                      "pin": self.otppin})
        r = token.authenticate("{}123456".format(self.otppin))
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], -1)
        # failure, but the token secret has been updated nonetheless
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Y"*224)
        # wrong PIN, the token secret has not been updated
        r = token.authenticate("WRONG123456".format(self.otppin))
        self.assertEqual(r[0], False)
        self.assertEqual(r[1], -1)
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Y"*224)
        # another failure, but the token secret has been updated again!
        r = token.authenticate("{}234567".format(self.otppin))
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], -1)
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Z"*224)
        token.delete_token()

    @mock_verification(mock_success)
    def test_05_success(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X"*248).decode("utf-8"),
                      "pin": self.otppin})
        # wrong PIN, the token secret has not been updated
        r = token.authenticate("WRONG123456".format(self.otppin))
        self.assertEqual(r[0], False)
        self.assertEqual(r[1], -1)
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"X"*224)
        # correct PIN + OTP
        r = token.authenticate("{}123456".format(self.otppin))
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], 0) # TODO: that is success?
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Y"*224)
        # another success
        r = token.authenticate("{}234567".format(self.otppin))
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], 0)  # TODO: that is success?
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Z"*224)
        token.delete_token()

    def test_06_reuse(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X" * 248).decode("utf-8"),
                      "pin": self.otppin})

        @mock_verification(mock_success)
        def _step1():
            # correct PIN + OTP
            return token.authenticate("{}123456".format(self.otppin))

        r = _step1()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], 0)  # TODO: that is success?
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X" * 24 + b"Y" * 224)

        # set another pin
        token.set_pin("anotherpin")

        @mock_verification(create_mock_failure(201))
        def _step2():
            # correct PIN, wrong OTP
            return token.authenticate("anotherpin123456")

        # reuse
        r = _step2()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], -1)
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X" * 24 + b"Z" * 224)

        # correct PIN + OTP
        @mock_verification(mock_success)
        def _step3():
            # correct PIN + OTP
            return token.authenticate("anotherpin234567")

        r = _step3()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], 0)
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X" * 24 + b"[" * 224)

        token.delete_token()

    def test_07_failures(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"X"*248).decode('utf-8'),
                      "pin": self.otppin})

        @mock_verification(create_mock_failure(123))
        def _step1():
            return token.authenticate("{}123456".format(self.otppin))

        r = _step1()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], -1)
        # failure, but the token secret has been updated nonetheless
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Y"*224)

        @mock_verification(create_mock_failure(202))
        def _step2():
            return token.authenticate("{}123456".format(self.otppin))
        r = _step2()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1], -1)
        # failure, but the token secret has been updated nonetheless
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"X"*24 + b"Z"*224)

        token.delete_token()

    def test_08_invalid_otpkey(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        self.assertRaises(ParameterError,
                          token.update,
                          {"otpkey": hexlify(b"X"*250).decode("utf-8")}) # wrong length
        self.assertRaises(ParameterError,
                          token.update,
                          {"otpkey": "X"*496}) # not a hex-string
        token.delete_token()

    def test_09_failcount(self):
        db_token = Token(self.serial2, tokentype="vasco")
        db_token.save()
        token = VascoTokenClass(db_token)
        token.update({"otpkey": hexlify(b"A" * 248).decode("utf-8"),
                      "pin": self.otppin})

        @mock_verification(create_mock_failure(1))
        def _step1():
            # correct PIN, wrong OTP
            return check_serial_pass(self.serial2, "{}123456".format(self.otppin))

        self.assertTrue(token.check_failcount())
        # fail 10 times
        for _ in range(10 + 1):
            r = _step1()
            self.assertEqual(r[0], False)
            self.assertEqual(r[1].get('message'), 'wrong otp value')

        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"A" * 24 + b"L" * 224)
        # fail counter has been exceeded
        self.assertFalse(token.check_failcount())

        @mock_verification(mock_success)
        def _step2():
            # correct PIN, wrong OTP
            return check_serial_pass(self.serial2, "{}123456".format(self.otppin))

        # subsequent authentication attempt fails due to fail counter
        r = _step2()
        self.assertEqual(r[0], False)
        self.assertEqual(r[1].get('message'), 'matching 1 tokens, Failcounter exceeded')
        # this actually does update the OTP key
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"A" * 24 + b"M" * 224)

        # reset the failcounter
        token.reset()

        # now, authentication works again
        r = _step2()
        self.assertEqual(r[0], True)
        self.assertEqual(r[1].get('message'), 'matching 1 tokens')
        key = token.token.get_otpkey().getKey()
        self.assertEqual(key, b"A" * 24 + b"N" * 224)

        token.delete_token()


""" Test module for a fancy token class """

from privacyidea.lib.tokenclass import TokenClass


class FancyTokenClass(TokenClass):
    def __init__(self, db_token):
        TokenClass.__init__(self, db_token)
        self.set_type("fancy")
        self.mode = ['authenticate']
        self.hKeyRequired = False

    @staticmethod
    def get_class_type():
        return 'fancy'

    @staticmethod
    def get_class_prefix():
        return "PIFA"



# http://www.privacyidea.org
# (c) Cornelius Kölbel, privacyidea.org
#
# 2016-01-01 Cornelius Kölbel <cornelius@privacyidea.org>
#            Password recovery
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__doc__ = """This module provides the REST API for the password recovery for a
user managed in privacyIDEA.

The methods are also tested in the file tests/test_api_register.py
"""
from flask import (Blueprint, request, g, current_app)
from .lib.utils import send_result, getParam
from .lib.utils import required
from privacyidea.lib.user import get_user_from_param
import logging
from privacyidea.lib.passwordreset import (create_recoverycode,
                                           check_recoverycode)
from privacyidea.lib.policy import ACTION
from privacyidea.api.lib.prepolicy import prepolicy, check_anonymous_user


log = logging.getLogger(__name__)

recover_blueprint = Blueprint('recover_blueprint', __name__)


# The before and after methods are the same as in the validate endpoint

@recover_blueprint.route('', methods=['POST'])
@prepolicy(check_anonymous_user, request, action=ACTION.PASSWORDRESET)
def get_recover_code():
    """
    This method requests a recover code for a user. The recover code it sent
    via email to the user.

    :queryparam user: username of the user
    :queryparam realm: realm of the user
    :queryparam email: email of the user
    :return: JSON with value=True or value=False
    """
    param = request.all_data
    user_obj = get_user_from_param(param, required)
    email = getParam(param, "email", required)
    r = create_recoverycode(user_obj, email, base_url=request.base_url)
    g.audit_object.log({"success": r,
                        "info": "{0!s}".format(user_obj)})
    return send_result(r)


@recover_blueprint.route('/reset', methods=['POST'])
@prepolicy(check_anonymous_user, request, action=ACTION.PASSWORDRESET)
def reset_password():
    """
    reset the password with a given recovery code.
    The recovery code was sent by get_recover_code and is bound to a certain
    user.

    :jsonparam recoverycode: The recoverycode sent the the user
    :jsonparam password: The new password of the user

    :return: a json result with a boolean "result": true
    """
    r = False
    user_obj = get_user_from_param(request.all_data, required)
    recoverycode = getParam(request.all_data, "recoverycode", required)
    password = getParam(request.all_data, "password", required)
    if check_recoverycode(user_obj, recoverycode):
        # set password
        r = user_obj.update_user_info({"password": password})
        g.audit_object.log({"success": r,
                            "info": "{0!s}".format(user_obj)})
    return send_result(r)



"""add timeout and retries to radiusserver table

Revision ID: 14a1bcb10018
Revises: 36428afb2457
Create Date: 2017-10-30 15:16:04.819246

"""

# revision identifiers, used by Alembic.
revision = '14a1bcb10018'
down_revision = '36428afb2457'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('radiusserver', sa.Column('retries', sa.Integer(), nullable=True))
        op.add_column('radiusserver', sa.Column('timeout', sa.Integer(), nullable=True))
    except Exception as exx:
        print("Could not add retries and timeout to radiusserver")
        print (exx)


def downgrade():
    op.drop_column('radiusserver', 'timeout')
    op.drop_column('radiusserver', 'retries')



# 2022-09-29 Cornelius Kölbel, <cornelius.koelbel@netknights.it>
#            Add new API for tokengroups
#
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """The tokengroup endpoint allows administrators to manage tokengroup definitions.
"""
from flask import (Blueprint, request)
from .lib.utils import (getParam, send_result)
from ..lib.log import log_with
from privacyidea.lib.tokengroup import get_tokengroups, set_tokengroup, delete_tokengroup
from privacyidea.lib.event import event
from privacyidea.lib.policy import ACTION
from privacyidea.api.lib.prepolicy import prepolicy, check_base_action

from flask import g
import logging

log = logging.getLogger(__name__)

tokengroup_blueprint = Blueprint('tokengroup_blueprint', __name__)


@tokengroup_blueprint.route('/<groupname>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.TOKENGROUP_ADD)
@event("tokengroup_add", request, g)
@log_with(log)
def set_tokengroup_api(groupname):
    """
    This call creates a new tokengroup or updates the description
    of an existing tokengroup.

    Note, that the identifier (name) of the tokengroup needs to be unique.

    :param tokengroup: The unique name of the tokengroup
    :param description: The description of the tokengroup
    :return:

    **Example request**:

    To create a new tokengroup "groupA" with a description call:

    .. sourcecode:: http

       POST /tokengroup/groupA HTTP/1.1
       Host: example.com
       Accept: application/json
       Content-Length: 26
       Content-Type: application/x-www-form-urlencoded

       description=My cool first tokengroup

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

       {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
                    "status": true,
                    "value": 1
          }
          "version": "privacyIDEA unknown"
       }

    """
    param = request.all_data
    description = getParam(param, "description", optional=True)

    r = set_tokengroup(groupname, description)

    g.audit_object.log({'success': r > 0, 'info':  groupname})
    return send_result(r)


@tokengroup_blueprint.route('/<groupname>', methods=['GET'])
@tokengroup_blueprint.route('/', methods=['GET'])
@prepolicy(check_base_action, request, ACTION.TOKENGROUP_LIST)
@event("tokengroup_list", request, g)
@log_with(log)
def get_tokengroup_api(groupname=None):
    """
    This call returns the information for the given tokengroup.
    If no groupname is specified, it returns a list of all defined tokengroups.

    :return: a json result with a list of tokengropups


    **Example request**:

    .. sourcecode:: http

       GET /tokengroup/ HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "gruppe1": {"description": "1st group"},
              "gruppe2": {"description": "2nd group"}
              }
            }
          },
          "version": "privacyIDEA unknown"
        }
    """
    tgs = get_tokengroups(name=groupname)
    g.audit_object.log({"success": True})
    r_tokengroups = {}
    for tg in tgs:
        r_tokengroups[tg.name] = {"description": tg.Description,
                                  "id": tg.id}

    return send_result(r_tokengroups)


@tokengroup_blueprint.route('/<groupname>', methods=['DELETE'])
@prepolicy(check_base_action, request, ACTION.TOKENGROUP_DELETE)
@event("tokengroup_delete", request, g)
@log_with(log)
def delete_tokengroup_api(groupname=None):
    """
    This call deletes the given tokengroup.

    :param groupname: The name of the token.

    :return: a json result with value=1 if deleting the realm was successful

    **Example request**:

    .. sourcecode:: http

       DELETE /tokengroup/gruppe1 HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": 1
          },
          "version": "privacyIDEA unknown"
        }

    """
    ret = delete_tokengroup(groupname)
    g.audit_object.log({"success": True,
                        "info": groupname})

    return send_result(1)



#  2018-23-09 Paul Lettich <paul.lettich@netknights.it>
#             Add decrease and reset counter actions
#  2018-02-28 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writeup
#
# License:  AGPLv3
# (c) 2018. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module for
increasing/decreasing/resetting database counters.
These counters can be used by rrdtool to monitor values and print time series of
certain parameters.
"""
from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib import _
from privacyidea.lib.counter import increase, decrease, reset
import logging
from privacyidea.lib.utils import is_true
import traceback

log = logging.getLogger(__name__)


class CounterEventHandler(BaseEventHandler):
    """
    An CounterEventHandler needs to return a list of actions, which it can handle.
    It also returns a list of allowed action and conditions
    It returns an identifier, which can be used in the eventhandling definitions
    """

    identifier = "Counter"
    description = "This event handler increases arbitrary counters in the database."

    @property
    def allowed_positions(cls):
        """
        This returns the allowed positions of the event handler definition.
        :return: list of allowed positions
        """
        return ["post", "pre"]

    @property
    def actions(cls):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        actions = {"increase_counter": {
            "counter_name": {
                "type": "str",
                "description": _("The identifier/key of the counter.")}
            },
            "decrease_counter": {
                "counter_name": {
                    "type": "str",
                    "description": _("The identifier/key of the counter.")},
                'allow_negative_values': {
                    "type": "bool",
                    "description": _("Don't stop counter if it reaches zero.")}
            },
            "reset_counter": {
                "counter_name": {
                    "type": "str",
                    "description": _("The identifier/key of the counter.")}
        }}
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        """
        ret = True
        #g = options.get("g")
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})
        counter_name = handler_options.get("counter_name")

        if action == "increase_counter":
            increase(counter_name)
            log.debug("Increased the counter {0!s}.".format(counter_name))
        elif action == "decrease_counter":
            allow_negative = is_true(handler_options.get("allow_negative_values", False))
            decrease(counter_name, allow_negative)
            log.debug("Decreased the counter {0!s}.".format(counter_name))
        elif action == "reset_counter":
            reset(counter_name)
            log.debug("Reset the counter {0!s} to 0.".format(counter_name))

        return ret



# http://www.privacyidea.org
# (c) cornelius kölbel, privacyidea.org
#
# 2019-09-04 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add decorators for event handling
# 2016-07-12 Cornelius Kölbel <cornelius@privacyidea.org>
#            Add decorator for checking realmadmin realms during userlist
# 2014-12-08 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Complete rewrite during flask migration
#            Try to provide REST API
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from flask import (Blueprint,
                   request)
from .lib.utils import (getParam,
                        send_result)
from ..api.lib.prepolicy import (prepolicy, check_base_action, realmadmin,
                                 check_custom_user_attributes)
from ..lib.policy import ACTION, get_allowed_custom_attributes
from privacyidea.api.auth import admin_required, user_required
from privacyidea.lib.user import create_user, User, is_attribute_at_all
from privacyidea.lib.event import event


from flask import (g)
from ..lib.user import get_user_list
import logging


log = logging.getLogger(__name__)


user_blueprint = Blueprint('user_blueprint', __name__)


@user_blueprint.route('/', methods=['GET'])
@prepolicy(realmadmin, request, ACTION.USERLIST)
@prepolicy(check_base_action, request, ACTION.USERLIST)
@user_required
@event("user_list", request, g)
def get_users():
    """
    list the users in a realm

    A normal user can call this endpoint and will get information about his
    own account.

    :param realm: a realm that contains several resolvers. Only show users
                  from this realm
    :param resolver: a distinct resolvername
    :param <searchexpr>: a search expression, that depends on the ResolverClass

    :return: json result with "result": true and the userlist in "value".

    **Example request**:

    .. sourcecode:: http

       GET /user/?realm=realm1 HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": [
              {
                "description": "Cornelius K\u00f6lbel,,+49 151 2960 1417,+49 561 3166797,cornelius.koelbel@netknights.it",
                "email": "cornelius.koelbel@netknights.it",
                "givenname": "Cornelius",
                "mobile": "+49 151 2960 1417",
                "phone": "+49 561 3166797",
                "surname": "K\u00f6lbel",
                "userid": "1009",
                "username": "cornelius",
                "resolver": "name-of-resolver"
              }
            ]
          },
          "version": "privacyIDEA unknown"
        }
    """
    realm = getParam(request.all_data, "realm")
    attr = is_attribute_at_all()
    users = get_user_list(request.all_data, custom_attributes=attr)

    g.audit_object.log({'success': True,
                        'info': "realm: {0!s}".format(realm)})

    return send_result(users)


@user_blueprint.route('/attribute', methods=['POST'])
@prepolicy(check_custom_user_attributes, request, "set")
@user_required
@event("set_custom_user_attribute", request, g)
def set_user_attribute():
    """
    Set a custom attribute for a user.
    The user is specified by the usual parameters user, resolver and realm.
    When a user is calling the endpoint the parameters will be implicitly set.

    :httpparam user: The username of the user, for whom the attribute should be set
    :httpparam resolver: The resolver of the user (optional)
    :httpparam realm: The realm of the user (optional)
    :httpparam key: The name of the attributes
    :httpparam value: The value of the attribute
    :httpparam type: an optional type of the attribute

    The database id of the attribute is returned. The return
    value thus should be >=0.
    """
    # We basically need a user, otherwise we will fail, but the
    # user object is later simply used from request.User. We only
    # need to avoid an empty User object.
    _user = getParam(request.all_data, "user", optional=False)
    attrkey = getParam(request.all_data, "key", optional=False)
    attrvalue = getParam(request.all_data, "value", optional=False)
    attrtype = getParam(request.all_data, "type", optional=True)
    r = request.User.set_attribute(attrkey, attrvalue, attrtype)
    g.audit_object.log({"success": True,
                        "info": "{0!s}".format(attrkey)})
    return send_result(r)


@user_blueprint.route('/attribute', methods=['GET'])
@user_required
@event("get_user_attribute", request, g)
def get_user_attribute():
    """
    Return the *custom* attribute of the given user.
    This does *not* return the user attributes which are contained in the user store!
    The user is specified by the usual parameters user, resolver and realm.
    When a user is calling the endpoint the parameters will be implicitly set.

    :httpparam user: The username of the user, for whom the attribute should be set
    :httpparam resolver: The resolver of the user (optional)
    :httpparam realm: The realm of the user (optional)
    :httpparam key: The optional name of the attribute. If it is not specified
         all custom attributes of the user are returned.

    """
    _user = getParam(request.all_data, "user", optional=False)
    attrkey = getParam(request.all_data, "key", optional=True)
    r = request.User.attributes
    if attrkey:
        r = r.get(attrkey)
    g.audit_object.log({"success": True,
                        "info": "{0!s}".format(attrkey)})
    return send_result(r)


@user_blueprint.route('/editable_attributes/', methods=['GET'])
@user_required
@event("get_editable_attributes", request, g)
def get_editable_attributes():
    """
    The resulting editable custom attributes according to the policies
    are returned. This can be a user specific result.
    When a user is calling the endpoint the parameters will be implicitly set.

    :httpparam user: The username of the user, for whom the attribute should be set
    :httpparam resolver: The resolver of the user (optional)
    :httpparam realm: The realm of the user (optional)

    Works for admins and normal users.
    :return:
    """
    _user = getParam(request.all_data, "user", optional=False)
    r = get_allowed_custom_attributes(g, request.User)
    g.audit_object.log({"success": True})
    return send_result(r)


@user_blueprint.route('/attribute/<attrkey>/<username>/<realm>', methods=['DELETE'])
@prepolicy(check_custom_user_attributes, request, "delete")
@user_required
@event("delete_custom_user_attribute", request, g)
def delete_user_attribute(attrkey, username, realm=None):
    """
    Delete a specified custom attribute from the user.
    The user is specified by the positional parameters user and realm.

    :httpparam user: The username of the user, for whom the attribute should be set
    :httpparam realm: The realm of the user
    :httpparam key: The name of the attribute that should be deleted from the user.

    Returns the number of deleted attributes.
    """
    user = User(username, realm)
    r = user.delete_attribute(attrkey)
    g.audit_object.log({"success": True,
                        "info": "{0!s}".format(attrkey)})
    return send_result(r)


@user_blueprint.route('/<resolvername>/<username>', methods=['DELETE'])
@admin_required
@prepolicy(check_base_action, request, ACTION.DELETEUSER)
@event("user_delete", request, g)
def delete_user(resolvername=None, username=None):
    """
    Delete a User in the user store.
    The resolver must have the flag editable, so that the user can be deleted.
    Only administrators are allowed to delete users.

    Delete a user object in a user store by calling

    **Example request**:

    .. sourcecode:: http

       DELETE /user/<resolvername>/<username>
       Host: example.com
       Accept: application/json

    """
    user_obj = request.User
    res = user_obj.delete()
    g.audit_object.log({"success": res,
                        "info": "{0!s}".format(user_obj)})
    return send_result(res)


@user_blueprint.route('', methods=['POST'])
@user_blueprint.route('/', methods=['POST'])
@admin_required
@prepolicy(check_base_action, request, ACTION.ADDUSER)
@event("user_add", request, g)
def create_user_api():
    """
    Create a new user in the given resolver.

    **Example request**:

    .. sourcecode:: http

       POST /user
       user=new_user
       resolver=<resolvername>
       surname=...
       givenname=...
       email=...
       mobile=...
       phone=...
       password=...
       description=...

       Host: example.com
       Accept: application/json

    """
    # We can not use "get_user_from_param", since this checks the existence
    # of the user.
    attributes = _get_attributes_from_param(request.all_data)
    username = getParam(request.all_data, "user", optional=False)
    resolvername = getParam(request.all_data, "resolver", optional=False)
    # Remove the password from the attributes, so that we can hide it in the
    # logs
    password = attributes.get("password")
    if "password" in attributes:
        del attributes["password"]
    r = create_user(resolvername, attributes, password=password)
    g.audit_object.log({"success": True,
                        "info": "{0!s}: {1!s}/{2!s}".format(r, username,
                                                            resolvername)})
    return send_result(r)


@user_blueprint.route('', methods=['PUT'])
@user_blueprint.route('/', methods=['PUT'])
@prepolicy(check_base_action, request, ACTION.UPDATEUSER)
@event("user_update", request, g)
def update_user():
    """
    Edit a user in the user store.
    The resolver must have the flag editable, so that the user can be deleted.
    Only administrators are allowed to edit users.

    **Example request**:

    .. sourcecode:: http

       PUT /user
       user=existing_user
       resolver=<resolvername>
       surname=...
       givenname=...
       email=...
       mobile=...
       phone=...
       password=...
       description=...

       Host: example.com
       Accept: application/json

    .. note:: Also a user can call this function to e.g. change his password.
       But in this case the parameter "user" and "resolver" get overwritten
       by the values of the authenticated user, even if he specifies another
       username.
    """
    attributes = _get_attributes_from_param(request.all_data)
    username = getParam(request.all_data, "user", optional=False)
    resolvername = getParam(request.all_data, "resolver", optional=False)
    userid = getParam(request.all_data, "userid")
    if userid is not None:
        # Create the user object by uid
        user_obj = User(resolver=resolvername, uid=userid)
    else:
        user_obj = User(login=username, resolver=resolvername)
    # Remove the password from the attributes, so that we can hide it in the
    # logs
    password = attributes.get("password")
    if password:
        del attributes["password"]
    r = user_obj.update_user_info(attributes, password=password)
    g.audit_object.log({"success": True,
                        "info": "{0!s}: {1!s}/{2!s}".format(r, username, resolvername)})
    return send_result(r)


def _get_attributes_from_param(param):
    from privacyidea.lib.resolver import get_resolver_object
    map = get_resolver_object(getParam(param, "resolver", optional=False)).map
    username = getParam(param, "user", optional=False)

    # Add attributes
    attributes = {"username": username}
    for attribute in map.keys():
        value = getParam(param, attribute)
        if value:
            attributes[attribute] = getParam(param, attribute)

    return attributes



"""Update AuthCache.authentication to VARCHAR(255) to accommodate argon2 hashes.

Revision ID: 8d40dbcfda25
Revises: 140ba0ca4f07
Create Date: 2020-11-02 17:57:12.493515

"""

# revision identifiers, used by Alembic.
revision = '8d40dbcfda25'
down_revision = '140ba0ca4f07'

from alembic import op
import sqlalchemy as sa

def upgrade():
    try:
        # ### commands auto generated by Alembic - please adjust! ###
        op.alter_column('authcache', 'authentication',
                        existing_type=sa.Unicode(length=64),
                        type_=sa.Unicode(length=255),
                        existing_nullable=True)
        # ### end Alembic commands ###
    except Exception as exx:
        print('Modifying of authcache column "authentication" failed: {!r}'.format(exx))
        print('This is expected behavior if the columns have already been modified.')


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('authcache', 'authentication',
                    existing_type=sa.Unicode(length=255),
                    type_=sa.Unicode(length=64),
                    existing_nullable=True)
    # ### end Alembic commands ###
